
/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * contains WebDocWf extensions
 *
 *-----------------------------------------------------------------------------
 * org.enhydra.barracuda.discRack.data.disc/DiscDO.java
 *-----------------------------------------------------------------------------
 */

package org.enhydra.barracuda.discRack.data.disc;

import java.io.*;
import java.sql.*;
import java.math.*;
import java.util.Hashtable;
import java.util.Collection;
import java.util.Properties;
import java.io.FileInputStream;

import org.enhydra.dods.DODS;
import com.lutris.util.Config;
import com.lutris.util.ConfigException;
import com.lutris.appserver.server.sql.*;
import com.lutris.dods.builder.generator.dataobject.*;
import com.lutris.dods.builder.generator.query.*;
import org.enhydra.dods.cache.LRUCache;
import org.enhydra.dods.cache.Condition;
import org.enhydra.dods.cache.QueryCacheImpl;
import org.enhydra.dods.cache.QueryCache;
import org.enhydra.dods.cache.QueryResult;
import org.enhydra.dods.cache.CacheAdministration;
import org.enhydra.dods.statistics.Statistics;
import org.enhydra.dods.statistics.TableStatistics;
import org.enhydra.dods.statistics.QueryCacheStatistics;

// WebDocWf extension for DODS row instance security and genric store
import org.webdocwf.dods.access.*;
import org.webdocwf.util.genericstore.GenericStoreDO;
import java.util.Vector;
import java.util.HashSet;
import java.util.Iterator;
// end of WebDocWf extension for DODS row instance security and genric store

/**
 * Data core class, used to set and retrieve the DiscDO information.
 *
 * @version $Revision: 1.7 $
 * @author  NN
 * @since   Enhydra Project
 */
 public class DiscDO extends com.lutris.dods.builder.generator.dataobject.GenericDO implements DiscDOI, java.io.Serializable {
    /**
     * Static final data members name the table and columns for this DO.
     * By using these members with an instance of the QueryBuilder class,
     * an application can perform straight SQL queries while retaining
     * compile-time checking of table and column usage.
     *
     * Example:  List the Cities containing Persons named Bob:
     *
     *    Using straight SQL with QueryBuilder:
     *      Pro: code runs faster because you create fewer objects
     *      Con: code is less clear
     *
     *         Vector fields = new Vector();
     *         fields.addElement( AddressDO.City );
     *         QueryBuilder qb = new QueryBuilder( fields );
     *         qb.addWhere( PersonDO.FirstName, "Bob" );
     *         qb.addWhere( PersonDO.PrimaryKey, AddressDO.Person );
     *         RDBRow row;
     *         while ( null != ( row = qb.getNextRow() ) ) {
     *             String city = row.get( AddressDO.City ).getString();
     *         }
     *
     *    Using Query/DO classes:
     *      Pro: code is (often) clearer
     *      Con: code runs slower because you create more objects
     *
     *         PersonQuery pq = new PersonQuery();
     *         pq.setQueryFirstName( "Bob" );
     *         PersonDO[] bobs = pq.getDOArray();
     *         for ( int i = 0; i < bobs.length; i++ ) {
     *             AddressQuery aq = new AddressQuery();
     *             aq.setQueryPerson( bobs[i] );
     *             AddressDO addr = aq.getNextDO();
     *             String city = addr.getCity();
     *         }
     */
    static public final RDBTable  table = new RDBTable( "Disc" );

    /**
     * Name of the logical database for which DO object was created
     */
    protected String originDatabase = null;

    /**
     * Return the name of the logical database for which DO object was created.
     *
     * @return origin logical database name.
     *
     */
    public String getOriginDatabase() {
        return originDatabase;
    }

    /**
     * Return Disc as the name of the table in the database
     * which contains DiscDO objects.
     * This method overrides CoreDO.getTableName()
     * and is used by CoreDO.executeUpdate() during error handling.
     *
     * @return The database table name.
     *
     * @see CoreDO
     * @author Jay Gunter
     */
    public String getTableName() {
        return "Disc";
    }

    /**
     * static final RDBColumn PrimaryKey for use with QueryBuilder.
     * See example above.
     */
    static public final RDBColumn PrimaryKey = new RDBColumn( table, getPrimaryKeyName() );

    /* RDBColumns for DiscDO attributes are defined below. */


    /* Using a DO (and its Query class) to access a VIEW instead of a TABLE:
     *
     * A DO (and its Query class) can be used to access a VIEW
     * instead of a TABLE.  The Data Object is created as usual in DODS,
     * but the "create table" SQL command for that DO is not used.
     * Instead, you substitute a "create view" command to create a
     * virtual table in the database; this is often done to provide
     * convenient access to a collection of tables joined together.
     *
     * A VIEW usually does not return "oid" and "version" columns;
     * often (but now always) a VIEW is defined to return the "oid" column
     * for one of the tables joined together in the definition of the VIEW.
     *
     * If notUsingOId is true, DiscDO.createExisting(ResultSet)
     * will NOT invoke the GenericDO(ResultSet) constructor
     * so to avoid attempting to extract the "oid" and "version" columns
     * from the ResultSet.
     */
    static protected final boolean notUsingOId = false;


    /**
     * A DO class contains a reference to a DataStruct class.
     * This reference can be null (when the data for the DO
     * has not yet been retrieved from the database),
     * allowing a DO object to be a lightweight placeholder
     * until its data is needed.
     */
    private DiscDataStruct data = null;

    /**
     * Return information whether the data for this object has been marked read-only.
     *
     * @return True if the data for this object has been marked read-only.
     *
     */
    public boolean isReadOnly() {
        if ( null == data )
            return false;
        return data.readOnly;
    }

    /**
     * Sets DO's data.
     */
     public void setData (Object data) {
        this.data = (DiscDataStruct)data;
     }

    /**
     * Returns DO's data.
     */
     public Object getData () {
        return data;
     }

    /**
     * Protected constructor. Only derived classes should call it.
     *
     * @param is_view
     *
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     */
    protected DiscDO ( boolean is_view )
    throws ObjectIdException, DatabaseManagerException {
        super( is_view );
        originDatabase = getLogicalDBName();
    }

    /**
     * Protected constructor.  Only derived classes should call it.
     *
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     */
    protected DiscDO ()
    throws ObjectIdException, DatabaseManagerException {
        super( notUsingOId );

        originDatabase = getLogicalDBName();
    }


    /**
     * isLoaded()
     * Returns information whether object's data is loaded from database.
     * @return true if the data for this object has been retrieved
     * from the database.
     */
    public boolean isLoaded() {
        return ( null != data );
    }

    /**
     * Load the fields for the DO from the database.
     *
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DataObjectException
     *   If the object is not found in the database.
     * @exception SQLException
     *   If the database rejects the SQL generated to retrieve data
     *   for this object, or if the table contains a bad foreign key, etc.
     */
    public void loadData()
    throws SQLException, ObjectIdException, DataObjectException {
        if ( null == data ) {
            data = new DiscDataStruct ();
        }

        ObjectId id = getOId();
        if ( null == id )
            return;
        if ( ! isPersistent() )   // DO from createVirgin
            return;

        // DO from createExisting.  Complain if no record in database.
        DiscQuery query;

                query = new DiscQuery ();

        query.setQueryOId( id );
        query.requireUniqueInstance();
        DiscDO obj;
        try {
            obj = query.getNextDO();
            if ( null == obj )
                throw new DataObjectException("DiscDO DO not found for id=" + id );
            makeIdentical(obj);
            setVersion(    obj.getVersion() );
            setNewVersion( obj.getVersion() );
        } catch ( NonUniqueQueryException e ) {
            throw new ObjectIdException( "Duplicate ObjectId" );
        }

    }

    /**
     * Load the actual DO data if necessary.
     * Called by get/set methods.
     *
     * @exception DataObjectException If a data access error occurs.
     */
    protected void checkLoad()
    throws DataObjectException {
        if ( null == data )
            try {
                loadData();
            } catch ( Exception e ) {
                throw new DataObjectException("Unable to load data for DiscDO id=" + getOId() +
                                              ", error = " + e.getMessage() );
            }
    }

    /**
     * Protected constructor used by createExisting(ObjectId) above.
     *
     * @param id The ObjectId for the object.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     * @exception SQLException
     *   Should never see this exception since GenericDO.ctor(ObjectId)
     *   never accesses the database.
     */
    protected DiscDO( ObjectId id )
    throws SQLException, ObjectIdException, DataObjectException, DatabaseManagerException {
        super( id );
        originDatabase = getLogicalDBName();
    }


    /**
     * Represents table and cache (if there is caching) statistics.
     */
    protected static Statistics statistics;

    /**
     * Get table statistics.
     *
     * @return Table statistics.
     */
    public static Statistics getStatistics() {
        if (cache != null) {
            ((QueryCacheStatistics)statistics).setCacheType(cache.getCacheType());
            ((QueryCacheStatistics)statistics).setMaxCacheSize(cache.getMaxCacheSize(false));
            ((QueryCacheStatistics)statistics).setCacheSize(cache.getCacheSize());
            ((QueryCacheStatistics)statistics).setComplexQueryCacheSize(cache.getComplexQueryCacheSize());
            ((QueryCacheStatistics)statistics).setMaxComplexQueryCacheSize(cache.getMaxComplexQueryCacheSize(false));
            ((QueryCacheStatistics)statistics).setSimpleQueryCacheSize(cache.getSimpleQueryCacheSize());
            ((QueryCacheStatistics)statistics).setMaxSimpleQueryCacheSize(cache.getMaxSimpleQueryCacheSize(false));
            statistics.stopTime();
        }
        return statistics;
    }

    /**
     * Refresh table statistics.
     */
    public static void refreshStatistics() {
        if (cache != null)
            statistics = new QueryCacheStatistics("Disc");
        else
            statistics = new TableStatistics("Disc");
    }

    /**
     * Get all used logical databases.
     *
     * @return Array that contains names of all used logical databases.
     *
     */
    public static String[] getUsedLogicalDatabases() {
        String[] str = { getLogicalDBName() };
        return str;
    }


    protected static QueryCacheImpl cache; // cache for DiscDO

    /**
     * Read cache configuration from application configuration file:
     * cache size for org.enhydra.barracuda.discRack.data.disc.Disc table or default cache size.
     */
    public static void readCacheConfiguration() {
        Config tableConfig = null;
        Config cacheConfig = null;
        int maxSize = 0;
        try {
            cacheConfig = (Config)DODS.getDatabaseManager().getConfig().getSection("DB.Cache");
            if (cacheConfig != null) {
                try {
                    tableConfig = cacheConfig.getConfig("Disc");
                } catch (ConfigException ex) {
                }
                boolean notFound = false;
                if (tableConfig != null) {
                    try {
                        maxSize = tableConfig.getInt("maxCacheSize");
                    } catch (ConfigException ex) {
                        notFound = true;
                    }
                } else
                    notFound = true;
                if ( notFound ) {
                    try {
                        maxSize = cacheConfig.getInt("defaultMaxCacheSize");
                        notFound = false;
                    } catch (ConfigException ex) {
                        notFound = true;
                    }
                }
                if ( notFound ) {
                    maxSize = QueryCache.DEFAULT_MAX_CACHE_SIZE;
                }
            } else
                maxSize = QueryCache.DEFAULT_MAX_CACHE_SIZE;

        } catch ( Exception e ) {
            maxSize = QueryCache.DEFAULT_MAX_CACHE_SIZE;
        }
        if (maxSize != 0) {
            if (cache == null) {
                cache = new QueryCacheImpl(maxSize);
            } else
                cache.setMaxCacheSize(maxSize);
            cache.readConfiguration(tableConfig,cacheConfig);
        } else
            cache = null;

//        System.out.println("Cache size for "+"org.enhydra.barracuda.discRack.data.disc.Disc"+" = "+getMaxCacheSize(true));
    }

    /**
     * Get name of the table that is cached.
     *
     * @return Name of the table that is cached.
     */
    public static String getCacheDodsTableName() {
        return "Disc";
    }

    /**
     * Get maximum cache size.
     *
     * @return Maximum cache size.
     */
    public static int getMaxCacheSize() {
        if (cache == null)
            return 0;
        return cache.getMaxCacheSize();
    }

    /**
     * Sets maximum cache size.
     *
     * @param maxSize Maximum cache size.
     */
    public static synchronized void setMaxCacheSize(int maxSize) {
        if (maxSize == 0) {
            cache = null;
            refreshStatistics();
            return;
        }
        if (maxSize != 0 && cache == null) {
            cache = new QueryCacheImpl(maxSize);
            try {
                refreshCache();
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else {
            cache.setMaxCacheSize(maxSize);
        }
    }

    /**
     * Returns size of currently used cache.
     *
     * @return size of currently used cache.
     */
    public static int getCacheSize() {
        if (cache != null)
            return cache.getCacheSize();
        return 0;
    }

    /**
     * Returns Disc table cache.
     *
     * @return Disc table cache.
     */
    public static CacheAdministration getCache() {
        return cache;
    }

    /**
     * Queries all rows in table, and for each row
     * creates a DO instance in the cache.
     * For these DOs, data.readOnly = true,
     * which causes set methods to throw an exception.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     * @exception SQLException
     *   If the database rejects the SQL generated to retrieve data
     *   for this object, or if the table contains a bad foreign key, etc.
     */
    public static void refreshCache()
    throws java.sql.SQLException, DatabaseManagerException, ObjectIdException, DataObjectException {
        if (cache != null) {
            cache.refresh();
            String querySnt = cache.getInitialQueryCache();
            if (querySnt != null) {
                DiscQuery query;

                    query = new DiscQuery ();
                    query.hitDatabase();
                    if (!querySnt.equals("*") ) {
                        QueryBuilder builder = query.getQueryBuilder();
                        builder.addWhere(querySnt);
                    }
                    DiscDO obj;
                    try {
                        if (cache.isFull())
                            while ( null != ( obj = query.getNextDO() ) )
                            obj.data.readOnly = true;
                        else
                            query.getNextDO();
                    } catch ( NonUniqueQueryException ex ) {
                        // Since we do not call query.requireUniqueInstance()
                        // this should never happen.
                        throw new ObjectIdException( "Duplicate ObjectId" );
                    }

            }
            refreshStatistics();
        }
    }

    /**
     * Add DO to cache.
     * If DO already exists in cache, just the data member is replaced,
     * so that application references to the DO remain valid.
     * If there  is no caching newDO object is returned.
     *
     * param newDO Data object that will be added to cache.
     *
     * return Data object added to cache.
     */
    public static synchronized DiscDO addToCache( DiscDO newDO ) {
        if (cache == null)
            return newDO;
        DiscDO ret = (DiscDO)cache.addDO(newDO);
        if (ret == null)
            return newDO;
        return ret;
    }

    /**
     * Add DO to cache.
     * If DO already exists in cache, just the data member is replaced,
     * so that application references to the DO remain valid.
     * This method overides method addToCache of the class CoreDO.
     */
    public void addToCache() {
        addToCache(this);
    }

    /**
     */
    public static synchronized DiscDO updateCache( DiscDO newDO ) {
        if (cache == null)
            return newDO;
        DiscDO ret = (DiscDO)cache.updateDO(newDO);
        if (ret == null)
            return newDO;
        return ret;
    }

    /**
     */
    public void updateCache() {
        updateCache(this);
    }

    /**
     */
    public static synchronized DiscDO deleteFromCache( DiscDO DO ) {
        if (cache != null)
            cache.deleteDO(DO);
        return DO;
    }
    /**
     * Remove DO from cache.
     *
     * @param dbName Logical name of the database from which
     * DiscDO object will be removed.
     * @param handle handle of Data object which will be re moved.
     */
    public static synchronized void removeFromCache(String dbName, String handle) {
        if (cache != null) {
            String cacheHandle = ""+handle;

            cache.removeDO( cacheHandle );
        }
    }

    /**
     */
    public void deleteFromCache() {
        deleteFromCache(this);
    }


    /**
     * Remove DO from cache.
     *
     * @param DO Data object which will be removed.
     *
     * @return Removed DO or null if DO object doesn't exist in the cache.
     */
    public static synchronized DiscDO removeFromCache( DiscDO DO ) {
        DiscDO oldDO = null;
        if (cache != null) {
            try {
                oldDO = DiscDO.createVirgin();
                oldDO.setOId(DO.getOId());
                oldDO.setVersion(DO.getVersion());
                oldDO.setData(DO.getData());
                cache.removeDO(DO);
            } catch (Exception e) {
            }
        }
        return oldDO;
    }

    /**
     * Remove DO from cache.
     * This method overides method evict() of the class CoreDO.
     */
    public void evict() {
        if (!isPersistent())
            removeFromCache(this);
    }

    /**
     * Remove DOs from cache.
     *
     * @param DOs Array of DOs which will be removed from cache.
     */
    public static void evict(DiscDO[] DOs) {
        for (int i=0; i<DOs.length; i++)
            DOs[i].evict();
    }

    /**
     * Remove DOs from cache.
     *
     * @param dbName Logical name of the database from which
     * DiscDOs object will be removed.
     * @param handles array of DOs handles that will be removed from cache.
     */
    public static void evict(String dbName, String[] handles) {
        if (handles!=null) {
            for (int i=0; i<handles.length; i++)
                removeFromCache(dbName, handles[i]);
        }
    }

    /**
     * Retrieve contents of the cache as an collection of DO objects.
     *
     * @return Collection of DO objects retrieved from the cache.
     */
    public static Collection getCacheContents() {
        if (cache!=null)
            return cache.getCache().values();
        return null;
    }

    /**
     * Get cache type.
     *
     * @return Cache type.
     */
    public static String getCacheType() {
        if (cache != null)
            return cache.getCacheType();
        return "none";
    }

    /**
     * The cache is used automatically by the class.
     * Callers of the create() methods do not know whether
     * the returned DO instance is shared or not.
     * This prevents the existence of multiple instances in memory
     * of the same database object.
     * The cache is instantiated when the first DO instance is created.
     */
    /**
     * Class that contains unchanging (static) data from the database
     * will have a cache of DOs representing the entire contents of the table.
     */
    static {
        try {
            refreshStatistics();
            readCacheConfiguration();
            refreshCache();
        } catch ( Exception e ) {
            // cannot throw from static block
        }
    }

    /**
     * This method is invoked whenever object needs to be loaded from database.
     *
     * @exception DataObjectException If a data access error occurs.
     */
    public void refresh() throws DataObjectException {
        try {
            loadData();
        } catch ( Exception e ) {
            throw new DataObjectException("Unable to load data for DiscDO id=" + getOId() +
                                          ", error = " + e.getMessage() );
        }
    }

    /**
     * This method is invoked whenever objects needs to be loaded from database.
     *
     * @param DOs Array of DOs which will be red from database.
     *
     * @exception DataObjectException If a data access error occurs.
     */
    public static void refresh(DiscDO[] DOs) throws DataObjectException {
        for (int i=0; i<DOs.length; i++)
            DOs[i].refresh();
    }

    /**
     * createVirgin()
     * Creates a DO that has no ObjectId or data.
     * Such a DO is used to insert a new database entry
     * after its data has been set.
     *
     * @return Created data object.
     *
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     */
    public static DiscDO createVirgin()
    throws DatabaseManagerException, ObjectIdException {
        return new DiscDO ();
    }


    /**
     * createExisting( BigDecimal )
     *
     * Factory method creates a DiscDO object by searching for it
     * in the database using the passed BigDecimal value as the primary key.
     *
     * Creates a DO that represents an existing entry in the database.
     * Such a DO is used to examine and possibly update such an entry.
     * createExisting() is called only from the code that retrieves
     * an ObjectId from a ResultSet (database query result).
     * createExisting() is protected because no other DO or BO should ever
     * need to call it.
     * FIX unfortunately the createExisting(BigDecimal) form *does* need
     * to be public because it is called by the public ctors of other DOs.
     * For example,
     *    AaaDO contains a ref to a BbbDO,
     *    so there is a method AaaDO.setBbb(BbbDO).
     *    In the ctor AaaDO(ResultSet), we have the call
     *       setBbb( BbbDO.createExisting( rs.getBigDecimal( "bbb", 0 )));
     * Since AaaDO is not in the same package as BbbDO,
     * BbbDO.createExisting(BigDecimal) must be public, not protected.
     * Java needs the C++ 'friend' idea.
     *
     * @param bd The BigDecimal representation of the ObjectId for the object.
     *
     * @return Created DiscDO object.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     * @exception SQLException
     *   If the database rejects the SQL generated to retrieve data
     *   for this object, or if the table contains a bad foreign key, etc.
     */
    public static DiscDO createExisting( BigDecimal bd )
    throws SQLException, ObjectIdException, DataObjectException, DatabaseManagerException {
        if ( null == bd )
            return null;
        return createExisting( new ObjectId( bd ) );
    }

    /**
     * createExisting( String )
     *
     * The createExisting method is used to create a <CODE>DiscDO</CODE>
     * from a string handle.
     *
     * @param handle String representation of the ObjectId for the object.
     *
     * @return Created DiscDO object.
     */
    public static DiscDO createExisting( String handle ) {
        DiscDO ret = null;
        try {
            BigDecimal bd = new BigDecimal( handle );
            ret = createExisting( bd );
        } catch ( Exception e ) {
        }
        return ret;
    }

    /**
     * createExisting( ObjectId )
     *
     * Factory method creates a DiscDO object by searching for it
     * in the database using the passed ObjectID value as the primary key.
     *
     * @param id The ObjectId for the object.
     *
     * @return Created DiscDO object.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     * @exception SQLException
     *   If the database rejects the SQL generated to retrieve data
     *   for this object, or if the table contains a bad foreign key, etc.
     */
    protected static DiscDO createExisting( ObjectId id )
    throws SQLException, ObjectIdException, DataObjectException, DatabaseManagerException {
        if ( null == id )
            return null;
        DiscDO ret = null;

        ret = new DiscDO( id );
        ret.setPersistent( true );  // mark DO as persistent (preexisting)
        if ( ! true ) { // If not lazy-loading, fetch DO data now.
            ret.loadData();
        } else {
            statistics.incrementLazyLoadingNum();
        }

        // unset the GenericDO.dirty flag.
        ret.markClean();
        return ret;
    }

    /**
     * createExisting( ResultSet )
     *
     * Factory method used to create an instance of this class to
     * represent a Data Object already existing in the database.
     *
     * @param rs The ResultSet returned by the Query class for
     * an existing Data Object stored in the database.
     *
     * @return Created DiscDO object.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     * @exception SQLException
     *   If the database rejects the SQL generated to retrieve data
     *   for this object, or if the table contains a bad foreign key, etc.
     */
    protected static DiscDO createExisting( ResultSet rs )
    throws SQLException, ObjectIdException, DataObjectException, DatabaseManagerException {
        if ( null == rs )
            return null;
        DiscDO ret = null;
        if ( notUsingOId ) {
            ret = new DiscDO ();
            ret.initFromResultSet( rs );
        } else {
            ret = new DiscDO ( rs );
        }
//        ret = addToCache( ret );
        return ret;
    }

    /**
     * createExisting( RDBRow )
     *
     * Factory method creates a DiscDO object by searching for it
     * in the database using the DiscDO.PrimaryKey value
     * in the passed RDBRow.
     *
     * @param RDBRow A row returned by QueryBuilder.getNextRow().
     *
     * @return Created DiscDO object.
     *
     * @exception DataObjectException
     *   If the RDBRow does not contain a DiscDO.PrimaryKey.
     *   If the object is not found in the database.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     * @exception SQLException
     *   If the database rejects the SQL generated to retrieve data
     *   for this object, or if the table contains a bad foreign key, etc.
     */
    protected static DiscDO createExisting( RDBRow row )
    throws SQLException, ObjectIdException, DataObjectException, DatabaseManagerException {
        if ( null == row )
            return null;
        RDBColumnValue pk = null;
        try {
            pk = row.get( DiscDO.PrimaryKey );
            return createExisting( pk );
        } catch ( Exception e ) {
            throw new DataObjectException("Cannot create DiscDO, row does not " +
                                          "contain DiscDO primary key." );
        }
    }

    /**
     * createExisting( RDBColumnValue )
     *
     * Factory method creates a DiscDO object by searching for it
     * in the database using the passed DiscDO.PrimaryKey.
     *
     * @param RDBColumnValue a PrimaryKey column value from a row
     * that was returned by QueryBuilder.getNextRow().
     *
     * @return Created DiscDO object.
     *
     * @exception DataObjectException
     *   If the RDBColumnValue does not contain a DiscDO.PrimaryKey.
     *   If the object is not found in the database.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     * @exception SQLException
     *   If the database rejects the SQL generated to retrieve data
     *   for this object, or if the table contains a bad foreign key, etc.
     */
    protected static DiscDO createExisting( RDBColumnValue pk )
    throws SQLException, ObjectIdException, DataObjectException, DatabaseManagerException {
        if ( null == pk )
            return null;
        if ( ! pk.equals( DiscDO.PrimaryKey ) )
            throw new DataObjectException("Cannot create DiscDO, " +
                                          "RDBColumnValue is not DiscDO.PrimaryKey." );
        BigDecimal bd = null;
        try {
            bd = pk.getBigDecimal();
        } catch ( Exception e ) {
            throw new DataObjectException("Cannot create DiscDO, bad primary key." );
        }
        if ( null == bd )
            return null;
        return createExisting( bd );
    }

    /**
     * Creates a DO that has no ObjectId
     * but has a copy of an existing DO's data.
     * Such a DO is used to insert a new database entry
     * that is largely similar to an existing entry.
     *
     * @param data The data struct to copy values from.
     *
     * @return Created DiscDO object.
     *
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     */
    public static DiscDO createCopy( DiscDataStruct data )
    throws DatabaseManagerException, ObjectIdException {
        DiscDO ret = new DiscDO ();
        ret.data = ( DiscDataStruct ) data.duplicate();
        return ret;
    }

    /**
     * Creates a DO that has no ObjectId
     * but has a copy of an existing DO's data.
     * Such a DO is used to insert a new database entry
     * that is largely similar to an existing entry.
     *
     * @param orig The original DO to copy.
     *
     * @return Created DiscDO object.
     *
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     */
    public static DiscDO createCopy( DiscDO orig )
    throws DatabaseManagerException, ObjectIdException {
        DiscDO ret = new DiscDO ();
        if ( null != orig.data ) {
            ret.data = ( DiscDataStruct ) orig.data.duplicate();
        }
        return ret;
    }

    /**
     * Causes the DO to refresh itself from the database
     * the next time a set or get method is called.
     */
    public void reload() {
        data = null;
    }

    /**
     * The methods <CODE>
     *     getHandle
     *     hasMatchingHandle

     *     findCachedObjectByHandle

     * </CODE> are used by Presentation Objects that need to populate
     * HTML select lists with Data Objects as options.
     * The <CODE>getHandle()</CODE> method is used
     * to set the value for each option,
     * and the <CODE>hasMatchingHandle()<CODE>

     * methods are used to lookup the Data Object when the selection has
     * been made.
     *
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     *
     * @return id of this DO as a string
     *   If an object's id can't be allocated for this object.
     */
    public String  getHandle()
    throws DatabaseManagerException {
        String ret = null;
        if ( null == getOId() )
               throw new DatabaseManagerException( "ID not set " );
        ret = getOId().toString();
        return ret;
    }

    /**
     * Returns cache handle.
     *
     * @return cache handle.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     */
    public String  getCacheHandle()
    throws DatabaseManagerException {
        String ret = getHandle();

        return ret;
    }

   /**
     * Compare string version of the id of this DO and handle.
     *
     * @param handle
     *   <CODE>String</CODE> version of DO id.
     *
     * @return boolean
     *   True if the string version of the id of this DO matches passed handle.
     *
     * @see getHandle
     */
    public boolean hasMatchingHandle( String handle ) {
        boolean ret = false;
        if ( null == getOId() )
           return false;
        else {
           String thisHnadle = getOId().toString();
           ret = thisHnadle.equals( handle );
        }
        return ret;
    }

   /**
     * Get data object with key cacheHandle from the cache.
     *
     * @param cacheHandle
     *   <CODE>String</CODE> version of

     * data object's id.
     *
     * @return <CODE>DiscDO</CODE>
     *   Object if one is found in cache, otherwise null.
     *
     * @see getHandle
     */
    public static DiscDO findCachedObjectByHandle( String cacheHandle ){
        if ( null == cacheHandle )
            return null;
        if ( null == cache )
            return null;
        return ( DiscDO ) cache.getDOByHandle( cacheHandle );
    }

    /**
     * Assigns the DataStruct of an existing DO to this DO.
     * Does not duplicate data. Just assigns the reference.
     *
     * @param orig The original DO.
     *
     */
    protected void makeIdentical( DiscDO orig ) {
        super.makeIdentical(orig);
        data = orig.data;
    }

    /**
     * getVersion makes the protected method public in CoreDO.
     *
     * @return Data object's version.
     */
    public int getVersion() {
        return super.getVersion();
    }

    // WebDocWf extension for writeable fully cached objects
    /**
     * Mark the object as read-only.
     *
     * WebDocWf extension
     *
     */
    public void makeReadOnly() {
        if ( null != data )
            data.readOnly=true;
    }

    /**
     * Mark the object as read-write.
     *
     * WebDocWf extension
     */
    public void makeReadWrite() {
        if ( null != data )
            data.readOnly=false;
    }
    // end of WebDocWf extension for writeable fully cached objects


    ////////////////////////// data member Title

    /**
     * static final RDBColumn Title for use with QueryBuilder.
     * See RDBColumn PrimaryKey at the top of this file for usage example.
     */
    static public final RDBColumn Title = new RDBColumn( table, "title", true );

    /**
     * Use for query caching.
     */
    static public final int COLUMN_TITLE = 0;
        static public final int title_MaxLength = 32;

    /**
     * Get title of the Disc.
     *
     * @return title of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    public String getTitle ()
    throws DataObjectException {
        checkLoad();
        return data.title;
    }
   
    /**
     * Set title of the Disc.
     *
     * @param title of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    public void setTitle ( String title )
    throws DataObjectException {
        _setTitle ( title );
    }

    /**
     * _setTitle is a protected method that is called by
     * setTitle if title is not part of
     * a multicolumn foreign key.
     *
     * @param title of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    protected void _setTitle ( String title )
    throws DataObjectException {
        checkLoad();
        data.title =  markNewValue(data.title, title , 0, title_MaxLength, false );
    }


    ////////////////////////// data member Artist

    /**
     * static final RDBColumn Artist for use with QueryBuilder.
     * See RDBColumn PrimaryKey at the top of this file for usage example.
     */
    static public final RDBColumn Artist = new RDBColumn( table, "artist", true );

    /**
     * Use for query caching.
     */
    static public final int COLUMN_ARTIST = 1;
        static public final int artist_MaxLength = 32;

    /**
     * Get artist of the Disc.
     *
     * @return artist of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    public String getArtist ()
    throws DataObjectException {
        checkLoad();
        return data.artist;
    }
   
    /**
     * Set artist of the Disc.
     *
     * @param artist of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    public void setArtist ( String artist )
    throws DataObjectException {
        _setArtist ( artist );
    }

    /**
     * _setArtist is a protected method that is called by
     * setArtist if artist is not part of
     * a multicolumn foreign key.
     *
     * @param artist of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    protected void _setArtist ( String artist )
    throws DataObjectException {
        checkLoad();
        data.artist =  markNewValue(data.artist, artist , 0, artist_MaxLength, false );
    }


    ////////////////////////// data member Genre

    /**
     * static final RDBColumn Genre for use with QueryBuilder.
     * See RDBColumn PrimaryKey at the top of this file for usage example.
     */
    static public final RDBColumn Genre = new RDBColumn( table, "genre", true );

    /**
     * Use for query caching.
     */
    static public final int COLUMN_GENRE = 2;
        static public final int genre_MaxLength = 32;

    /**
     * Get genre of the Disc.
     *
     * @return genre of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    public String getGenre ()
    throws DataObjectException {
        checkLoad();
        return data.genre;
    }
   
    /**
     * Set genre of the Disc.
     *
     * @param genre of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    public void setGenre ( String genre )
    throws DataObjectException {
        _setGenre ( genre );
    }

    /**
     * _setGenre is a protected method that is called by
     * setGenre if genre is not part of
     * a multicolumn foreign key.
     *
     * @param genre of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    protected void _setGenre ( String genre )
    throws DataObjectException {
        checkLoad();
        data.genre =  markNewValue(data.genre, genre , 0, genre_MaxLength, false );
    }


    ////////////////////////// data member Owner

    /**
     * static final RDBColumn Owner for use with QueryBuilder.
     * See RDBColumn PrimaryKey at the top of this file for usage example.
     */
    static public final RDBColumn Owner = new RDBColumn( table, "owner", true );

    /**
     * Use for query caching.
     */
    static public final int COLUMN_OWNER = 3;
        

    /**
     * Get owner of the Disc.
     *
     * @return owner of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    public org.enhydra.barracuda.discRack.data.person.PersonDO getOwner ()
    throws DataObjectException {
        checkLoad();
        return data.owner;
    }
   
    /**
     * Set owner of the Disc.
     *
     * @param owner of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    public void setOwner ( org.enhydra.barracuda.discRack.data.person.PersonDO owner )
    throws DataObjectException {
        _setOwner ( owner );
    }

    /**
     * _setOwner is a protected method that is called by
     * setOwner if owner is not part of
     * a multicolumn foreign key.
     *
     * @param owner of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    protected void _setOwner ( org.enhydra.barracuda.discRack.data.person.PersonDO owner )
    throws DataObjectException {
        checkLoad();
        data.owner = (org.enhydra.barracuda.discRack.data.person.PersonDO) markNewValue(data.owner, owner  );
    }


    ////////////////////////// data member IsLiked

    /**
     * static final RDBColumn IsLiked for use with QueryBuilder.
     * See RDBColumn PrimaryKey at the top of this file for usage example.
     */
    static public final RDBColumn IsLiked = new RDBColumn( table, "isLiked", true );

    /**
     * Use for query caching.
     */
    static public final int COLUMN_ISLIKED = 4;
        

    /**
     * Get isLiked of the Disc.
     *
     * @return isLiked of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    public boolean getIsLiked ()
    throws DataObjectException {
        checkLoad();
        return data.isLiked;
    }
   
    /**
     * Set isLiked of the Disc.
     *
     * @param isLiked of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    public void setIsLiked ( boolean isLiked )
    throws DataObjectException {
        _setIsLiked ( isLiked );
    }

    /**
     * _setIsLiked is a protected method that is called by
     * setIsLiked if isLiked is not part of
     * a multicolumn foreign key.
     *
     * @param isLiked of the Disc.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     */
    protected void _setIsLiked ( boolean isLiked )
    throws DataObjectException {
        checkLoad();
        data.isLiked =  markNewValue(data.isLiked, isLiked  );
    }

    // array of function objects 
    private static Compare comparators[] = new Compare[5];

    // Initialization function objects
    static {
        comparators[COLUMN_TITLE] = new Compare() {
            public boolean compare (GenericDO DO, Condition cond) {
                try {
                    return QueryBuilder.compare(((DiscDO)DO).getTitle(),cond.getValue(),cond.getOperator());
                } catch (Exception e) {}
                return false;
            }
        };
        comparators[COLUMN_ARTIST] = new Compare() {
            public boolean compare (GenericDO DO, Condition cond) {
                try {
                    return QueryBuilder.compare(((DiscDO)DO).getArtist(),cond.getValue(),cond.getOperator());
                } catch (Exception e) {}
                return false;
            }
        };
        comparators[COLUMN_GENRE] = new Compare() {
            public boolean compare (GenericDO DO, Condition cond) {
                try {
                    return QueryBuilder.compare(((DiscDO)DO).getGenre(),cond.getValue(),cond.getOperator());
                } catch (Exception e) {}
                return false;
            }
        };
        comparators[COLUMN_OWNER] = new Compare() {
            public boolean compare (GenericDO DO, Condition cond) {
                try {
                    return QueryBuilder.compare(((DiscDO)DO).getOwner(),cond.getValue(),cond.getOperator());
                } catch (Exception e) {}
                return false;
            }
        };
        comparators[COLUMN_ISLIKED] = new Compare() {
            public boolean compare (GenericDO DO, Condition cond) {
                try {
                    return QueryBuilder.compare(((DiscDO)DO).getIsLiked(),cond.getValue(),cond.getOperator());
                } catch (Exception e) {}
                return false;
            }
        };
    }
    /**
     * Compares whether this DO satisfies condition cond.
     *
     * @param cond condition.
     * 
     * @return true if DO satisfies condition cond, otherwise false.
     */
    public boolean compareCond(Condition cond) {
        try {
            return comparators[cond.getColumnIndex()].compare(this,cond);
        } catch (Exception e) {
        }
        return false;
    }

    /**
     * logicalDbName is logical database name
     * set by setLogicalDBName()
     * and retrieved by getLogicalDBName().
     */
    static private String logicalDbName = null;

    /**
     * setLogicalDBName sets the logical database name that will be used
     * to create DBTransaction and DBQuery objects used by
     * DiscDO and the corresponding Query class.
     * 
     * @param logicalDbNameInConfFile The logical database specified in the
     * application's .conf file.
     *
     * @deprecated It is dangeruous to use this method in multiuser environment because, 
     * this setings are applied to all users (sets logical database to all users)
     */
    static public synchronized void setLogicalDBName( String logicalDbNameInConfFile ) {
        if ( null != logicalDbNameInConfFile && 0    != logicalDbNameInConfFile.length() )
            logicalDbName = logicalDbNameInConfFile;
        else 
            logicalDbName = DODS.getDatabaseManager().getDefaultDB();
    }
    
    /**
     * getLogicalDBName retrieves the logical database name 
     * set by setLogicalDBName().
     *
     * @return the logical database name that was set by method setLogicalDBName()
     *
     * @deprecated 
     */
    static public synchronized String getLogicalDBName() {
        if (logicalDbName == null)
            logicalDbName = DODS.getDatabaseManager().getDefaultDB();                    
        return logicalDbName;
    }

    /**
     * createTransaction() creates a new DBTransaction.
     * This method uses the logical database name set by method setLogicalDBName().
     *
     * If setLogicalDBName() was used to set the logical database name
     * to something other than the value of DatabaseManager.DefaultDatabase
     * in the application's .conf file, then any DBTransaction passed to 
     * save(DBTransaction) or delete(DBTransaction) should be created using 
     * DiscDO.createTransaction().
     *
     * The DiscDO save() and delete() methods use this method.
     *
     * @return A DBTransaction object to use with the DiscDO class.
     */
    static public DBTransaction createTransaction() 
    throws DatabaseManagerException, SQLException {
        DBTransaction ret;
        try {
            ret = DODS.getDatabaseManager().createTransaction(getLogicalDBName());
            ret.setDatabaseName(getLogicalDBName());
            return ret;
        } catch ( DatabaseManagerException e ) { 
            String err = "";
            if ( null != getLogicalDBName() )
                err = "ERROR: Could not create a DBTransaction.  " +
                      "DiscDO.logicalDbName='" + getLogicalDBName() + "'.  "+
                      "The application .conf file must list this name in " +
                      "DatabaseManager.Databases[], and there must be " +
                      "DatabaseManager.DB." + getLogicalDBName() + " settings.";
            throw new DatabaseManagerException( err, e );
        }
    }

    /**
     * createQuery() creates a new DBQuery.
     * This method uses the logical database name set by method setLogicalDBName().
     *
     * If setLogicalDBName() was used to set the logical database name
     * to something other than the value of DatabaseManager.DefaultDatabase
     * in the application's .conf file, then any DBQuery object used to 
     * access the 'Disc' table should be created using 
     * DiscDO.createQuery().
     *
     * The Query class corresponding to DiscDO uses this method.
     *
     * @return A DBQuery object to use in accessing the 'Disc' table.
     */
    static public DBQuery createQuery() 
    throws DatabaseManagerException, SQLException {
        try {
            return DODS.getDatabaseManager().createQuery(getLogicalDBName());
        } catch ( DatabaseManagerException e ) { 
            String err = "";
            if ( null != getLogicalDBName() )
                err = "ERROR: Could not create a DBQuery.  " +
                      "DiscDO.logicalDBName='" + getLogicalDBName() + "'.  "+
                      "The application .conf file must list this name in " +
                      "DatabaseManager.Databases[], and there must be " +
                      "DatabaseManager.DB." + getLogicalDBName() + " settings.";
            throw new DatabaseManagerException( err, e );
        }
    }

    /**
     * Protected constructor.
     *
     * @param rs Result set from which to obtain product data.
     *
     * @exception DataObjectException
     *   If the object is not found in the database.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an object's id can't be allocated for this object.
     * @exception DatabaseManagerException
     *   If a connection to the database cannot be established, etc.
     * @exception SQLException
     *   If the database rejects the SQL generated to retrieve data
     *   for this object, or if the table contains a bad foreign key, etc.
     */
    protected DiscDO(ResultSet rs)
    throws SQLException, ObjectIdException, DataObjectException, DatabaseManagerException {
        super(rs);
        initFromResultSet( rs );
        originDatabase = getLogicalDBName();
    }

    /**
     * initFromResultSet initializes the data members of Disc.
     * This code was separated from the ResultSet constructor
     * so that createExisting(ResultSet) could handle VIEWs.
     *
     * @param rs ResultSet from which data members are initialized.
     */
    private void initFromResultSet( ResultSet rs )
    throws SQLException, ObjectIdException, DataObjectException, DatabaseManagerException {
        // Constructing a DO from a ResultSet means we definitely need the 
        // DataStruct ready for the setXxx methods invoked below.
        data = new DiscDataStruct ();
 
        // writeMemberStuff uses the ResultSetExtraction.template
        // to build up the value for this tag:
        // the value is a series of calls to the DO set methods.

            setTitle(  rs.getString( "title"  ) );

            setArtist(  rs.getString( "artist"  ) );

            setGenre(  rs.getString( "genre"  ) );

            setOwner( org.enhydra.barracuda.discRack.data.person.PersonDO.createExisting( rs.getBigDecimal( "owner", 0) ) );

            setIsLiked(  rs.getBoolean( "isLiked"  ) );

        markClean();
    }        


    private int[] param = null;

    /**
     * Prepares the statement used to insert this object
     * into the database.
     *
     * @param conn The database connection.
     *
     * @return The insert statement.
     *
     * @exception java.sql.SQLException if an error occurs.
     */
    public PreparedStatement getInsertStatement(DBConnection conn)
    throws SQLException {

        ObjectId oid;

        PreparedStatement stmt = conn.prepareStatement(  
            "insert into Disc ( title, artist, genre, owner, isLiked, " + getOIdColumnName() + ", " + getVersionColumnName() + " )" +
            "values ( ?, ?, ?, ?, ?, ?, ? )" 
        );

        param = new int[1]; param[0] = 1;
        // writeMemberStuff uses the JDBCsetCalls.template
        // to build up the value for this tag:
        // the value is a series of calls to setPrepStmtParam_TYPE methods.
        // Those methods are defined in GenericDO.
        try {
            setPrepStmtParam_String ( stmt, param, getTitle() ); 
            setPrepStmtParam_String ( stmt, param, getArtist() ); 
            setPrepStmtParam_String ( stmt, param, getGenre() ); 
            setPrepStmtParam_DO ( stmt, param, getOwner() ); 
            setPrepStmtParam_boolean ( stmt, param, getIsLiked() ); 

            /* The ObjecId/Version columns form the primary key.  */
            setPrepStmtParam_BigDecimal( stmt, param, getOId().toBigDecimal() );
            setPrepStmtParam_int( stmt, param, getNewVersion() );

        } catch ( Exception e ) {
            throw new SQLException( "Data Object error: " + e.getMessage() );
        }
        statistics.incrementInsertNum();
        return stmt;
    }

    /**
     * Prepares the statement used to update this object
     * in the database.
     *
     * @param conn The database connection
     *
     * @return The update statement.
     *
     * @exception java.sql.SQLException if an error occurs.
     */
    public PreparedStatement getUpdateStatement(DBConnection conn)
    throws SQLException {

        PreparedStatement stmt = conn.prepareStatement(
            "Update Disc set " + 
            getVersionColumnName() + " = ?, title = ?, artist = ?, genre = ?, owner = ?, isLiked = ? " + 
            "where " + getOIdColumnName() + " = ? and " + getVersionColumnName() + " = ?"
        );

        param = new int[1]; param[0] = 1;
        try {
            setPrepStmtParam_int( stmt, param, getNewVersion() );
            setPrepStmtParam_String( stmt, param, getTitle() );
            setPrepStmtParam_String( stmt, param, getArtist() );
            setPrepStmtParam_String( stmt, param, getGenre() );
            setPrepStmtParam_DO( stmt, param, getOwner() );
            setPrepStmtParam_boolean( stmt, param, getIsLiked() );
            setPrepStmtParam_BigDecimal( stmt, param, getOId().toBigDecimal() );
            setPrepStmtParam_int(        stmt, param, getVersion() );
        } catch ( Exception e ) {
            throw new SQLException( "Data Object error: " + e.getMessage() );
        }
        statistics.incrementUpdateNum();
        return stmt;
    }

    /**
     * Prepares the statement used to delete this object
     * from the database.
     *
     * @param conn The database connection
     *
     * @return The delete statement.
     *
     * @exception java.sql.SQLException if an error occurs.
     */
    public PreparedStatement getDeleteStatement(DBConnection conn)
    throws SQLException {
        String sql =
            "delete from Disc \n" +
            "where " + getOIdColumnName() + " = ?";
        PreparedStatement stmt = conn.prepareStatement(sql);
        stmt.setBigDecimal(1, getOId().toBigDecimal());
        statistics.incrementDeleteNum();
        return stmt;
    }


    /*
     * toString - for debugging
     *
     * @return String for debugging.
     *
     */
    public String toString(){
        return toString( 1 );
    }

    public String toString( int indentCount ){
        String indent = "";
        for ( int i = 0; i < indentCount; i++ ) {
            indent += ". ";
        }
        StringBuffer sb = new StringBuffer();
        sb.append(indent + "DiscDO:");

        ObjectId oid = getOId();
        String id = "virgin";
        if ( null != oid )
            id = oid.toString();
        sb.append(" OID=" + id + ",VERSION=" + getVersion());
        if ( null != data ) {    
                sb.append("\n" + indent + "title=" + data.title);
    
                sb.append("\n" + indent + "artist=" + data.artist);
    
                sb.append("\n" + indent + "genre=" + data.genre);
    
                sb.append("\n" + indent + "owner=" + ( null == data.owner ? null  : data.owner.toString( indentCount + 1 ) ));
    
                sb.append("\n" + indent + "isLiked=" + data.isLiked);
;
            sb.append("\n" + indent + "SUPER=" + super.toString( indentCount ));
        }
        return sb.toString(); 
    }

    /**
     * A stub method for implementing pre-commit assertions
     * for this DiscDO.
     * Implement this stub to throw an RefAssertionException for cases
     * where this object is not valid for writing to the database.
     */
    protected void okToCommit()
    throws RefAssertionException { }

    /**
     * A stub method for implementing pre-delete assertions
     * for this DiscDO.
     * Implement this stub to throw an RefAssertionException for cases
     * where this object is not valid for deletion from the database.
     */
    protected void okToDelete()
    throws RefAssertionException { }

    /**
     * Inserts/Updates the DO into its table.
     *
     * @exception com.lutris.appserver.server.sql.DatabaseManagerException If a Transaction can not be created.
     * @exception RefAssertionException Thrown by okTo method.
     * @exception java.sql.SQLException If any SQL errors occur.
     *
     * @deprecated Use save() instead.
     */
    public void commit()
    throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
        commit(null);
    }

    /**
     * Inserts/Updates the DO into its table.
     * The transaction is likely provided by the commit() method of another DO
     * which references this DO.
     *
     * @param dbt The transaction object used for this operation.
     *
     * @exception com.lutris.appserver.server.sql.DatabaseManagerException If a Transaction can not be created.
     * @exception com.lutris.appserver.server.sql.DBRowUpdateException If a version error occurs.
     * @exception RefAssertionException Thrown by okTo method.
     * @exception java.sql.SQLException If any SQL errors occur.
     *
     * @deprecated Use save() instead.
     */
    public void commit(DBTransaction dbt)
    throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
        if (cache != null)
            if (cache.getCacheReadOnly())
                throw new DataObjectException("DiscDO's cache is read-only. Therefore, DML opertions are not allowed.");
        // WebDocWf extension for generic store
        // The following line has been inserted:
        try {
            DBTransaction dbtlocal = dbt;
            boolean needToCommit = false;
            if (dbtlocal == null) {
                dbtlocal = DODS.getDatabaseManager().createTransaction(getOriginDatabase());
                dbtlocal.setDatabaseName(getOriginDatabase());
                needToCommit = true;
            }
            // The following line has been changed:
            modifyDO( dbtlocal, false );
            // before: modifyDO( dbt, false );
            if (needToCommit) {
                dbtlocal.commit();
                dbtlocal.release();
            }
        } catch (DataObjectException e) {
            modifyDO( dbt, false );
        }
        // end of WebDocWf extension for generic store
    }

    /**
     * Inserts/Updates the DO into its table.
     *
     * @exception com.lutris.appserver.server.sql.DatabaseManagerException If a Transaction can not be created.
     * @exception RefAssertionException Thrown by okTo method.
     * @exception java.sql.SQLException If any SQL errors occur.
     *
     * WebDocWf extension
     */
    public void save()
    throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
        save(null,true);
    }

    /**
     * Inserts/Updates the DO into its table.
     *
     * @param references True if references should be saved with this DO.
     *
     * @exception com.lutris.appserver.server.sql.DatabaseManagerException If a Transaction can not be created.
     * @exception RefAssertionException Thrown by okTo method.
     * @exception java.sql.SQLException If any SQL errors occur.
     *
     * WebDocWf extension
     */
    public void save(boolean references)
    throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
        save(null,references);
    }

    /**
     * Inserts/Updates the DO into its table.
     * The transaction is likely provided by the commit() method of another DO
     * which references this DO.
     *
     * @param dbt The transaction object used for this operation.
     *
     * @exception com.lutris.appserver.server.sql.DatabaseManagerException If a Transaction can not be created.
     * @exception com.lutris.appserver.server.sql.DBRowUpdateException If a version error occurs.
     * @exception RefAssertionException Thrown by okTo method.
     * @exception java.sql.SQLException If any SQL errors occur.
     *
     * WebDocWf extension
     */
    public void save(DBTransaction dbt)
    throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
        save(dbt, true);
    }

    /**
     * Inserts/Updates the DO into its table.
     * The transaction is likely provided by the commit() method of another DO
     * which references this DO.
     *
     * @param dbt The transaction object used for this operation.
     * @param references True if references of this DO should be saved.
     *
     * @exception com.lutris.appserver.server.sql.DatabaseManagerException If a Transaction can not be created.
     * @exception com.lutris.appserver.server.sql.DBRowUpdateException If a version error occurs.
     * @exception RefAssertionException Thrown by okTo method.
     * @exception java.sql.SQLException If any SQL errors occur.
     *
     * WebDocWf extension
     */
    public void save(DBTransaction dbt, boolean references)
    throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
        if (cache != null)
            if (cache.getCacheReadOnly())
                throw new DataObjectException("DiscDO's cache is read-only. Therefore, DML opertions are not allowed.");
        // before: modifyDO( dbt, false );
        // The following line has been inserted:
        try {  // Added by Nenad Vico 25.11.2002
            // WebDocWf extension for generic store
            // The following line has been inserted:
            DBTransaction dbtlocal = dbt;
            boolean needToCommit = false;
            if (dbtlocal == null) {
                dbtlocal = DODS.getDatabaseManager().createTransaction(getOriginDatabase());
                dbtlocal.setDatabaseName(getOriginDatabase());
                needToCommit = true;
            }
            // The following line has been changed:
            modifyDO( dbtlocal, false, references );
            if (needToCommit) {
                dbtlocal.commit();
                dbtlocal.release();
            }
        } catch (DataObjectException e) {  // Added by Nenad Vico 25.11.2002
            modifyDO( dbt, false );  // Added by Nenad Vico 25.11.2002
        }  // Added by Nenad Vico 25.11.2002
        // end of WebDocWf extension for generic store
    }

    /**
     * Deletes the DO from its table.
     *
     * @exception com.lutris.appserver.server.sql.DatabaseManagerException If a Transaction can not be created.
     * @exception RefAssertionException Thrown by okTo method.
     * @exception java.sql.SQLException If any SQL errors occur.
     */
    public void delete()
    throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
        delete((DBTransaction)null);
    }

    /**
     * Deletes the DO from its table.
     * The transaction is likely provided by the delete() method of another DO
     * which references this DO.
     *
     * @param dbt The transaction object used for this operation.
     *
     * @exception com.lutris.appserver.server.sql.DatabaseManagerException If a Transaction can not be created.
     * @exception com.lutris.appserver.server.sql.DBRowUpdateException If a version error occurs.
     * @exception RefAssertionException Thrown by okTo method.
     * @exception java.sql.SQLException If any SQL errors occur.
     */
    public void delete(DBTransaction dbt)
    throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {

        if (cache != null)
            if (cache.getCacheReadOnly())
                throw new DataObjectException("DiscDO's cache is read-only. Therefore, DML opertions are not allowed.");
        // WebDocWf extension for generic store
        // The following lines have been inserted:
        try {       // Added by Nenad Vico 25.11.2002
            DBTransaction dbtlocal = dbt;
            boolean needToCommit = false;
            if (dbtlocal == null) {
                dbtlocal = DODS.getDatabaseManager().createTransaction(getOriginDatabase());
                dbtlocal.setDatabaseName(getOriginDatabase());
                needToCommit = true;
            }
            // The following line has been changed:
            modifyDO( dbtlocal, true );
            // before:  modifyDO( dbt, true );
            // end of WebDocWf extension for generic store
            if (needToCommit) {
                dbtlocal.commit();
                dbtlocal.release();
            }
        } catch (DataObjectException e) { // Added by Nenad Vico 25.11.2002
            modifyDO( dbt, true ); // Added by Nenad Vico 25.11.2002
        } // Added by Nenad Vico 25.11.2002
    }
    
    /**
     * A stub method for implementing pre-commit assertions 
     * for the owner data member.
     * Implement this stub to throw an RefAssertionException for cases
     * where owner is not valid for writing to the database.
     *
     * @param member owner data member. 
     *
     */
    protected void okToCommitOwner( org.enhydra.barracuda.discRack.data.person.PersonDO member ) 
    throws RefAssertionException { }

    /**
     * A stub method for implementing pre-delete assertions 
     * for the owner data member.
     * Implement this stub to throw an RefAssertionException for cases
     * where owner is not valid for deletion from the database.
     *
     *@param member owner data member
     *
     */
    protected void okToDeleteOwner( org.enhydra.barracuda.discRack.data.person.PersonDO member ) 
    throws RefAssertionException { }

    /**
     * Modifies the DO within its table.
     * Performs recursive commit/delete on referenced DOs;
     * all operations occur within a single transaction
     * to allow rollback in the event of error.
     * Only the creator of the transaction releases it.
     *
     * @param dbt The transaction object used for this operation.
     * @param delete True if doing a delete, otherwise false (for insert/update).
     *
     * @exception com.lutris.appserver.server.sql.DatabaseManagerException If a Transaction can not be created.
     * @exception com.lutris.appserver.server.sql.DBRowUpdateException If a version error occurs.
     * @exception RefAssertionException Thrown by okTo method.
     * @exception java.sql.SQLException If any SQL errors occur.
     * 
     * WebDocWf extension
     */
    protected void modifyDO( DBTransaction dbt, boolean delete)
    throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
        modifyDO( dbt, delete, true);
    }

    /**
     * Modifies the DO within its table.
     * Performs recursive commit/delete on referenced DOs;
     * all operations occur within a single transaction
     * to allow rollback in the event of error.
     * Only the creator of the transaction releases it.
     *
     * @param dbt The transaction object used for this operation.
     * @param delete True if doing a delete, otherwise false (for insert/update).
     * @param references True if references should be saved
     *
     * @exception com.lutris.appserver.server.sql.DatabaseManagerException If a Transaction can not be created.
     * @exception com.lutris.appserver.server.sql.DBRowUpdateException If a version error occurs.
     * @exception RefAssertionException Thrown by okTo method.
     * @exception java.sql.SQLException If any SQL errors occur.
     * 
     * WebDocWf extension
     */
    protected void modifyDO( DBTransaction dbt, boolean delete, boolean references )
    throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
        if ( delete )
            okToDelete();
        else
            okToCommit();
        boolean ownTransaction = false;
        try {
            if ( null == dbt ) {
                DatabaseManager dbm = DODS.getDatabaseManager();
                dbt = dbm.createTransaction(getOriginDatabase());      // create a transaction
                dbt.setDatabaseName(getOriginDatabase());
                ownTransaction = true;
            } else {

                // WebDocWf fix for circular references
                // The following lines have been inserted
                CoreDO foundDO = dbt.getDO(this);
                if (foundDO!=null) 
                    return;
            }
            // end of WebDocWf fix for circular references
            if ( null == dbt )
                throw new DatabaseManagerException("DatabaseManager.createTransaction returned null." );
            if ( delete ) {
                // Code to perform cascading deletes is generated here
                // if cascading deletes are not supported by the database.      

                // The following line keeps the compiler happy 
                // when the CASCADING_DELETES tag is empty.
                if ( false )
                    throw new QueryException("XXX");
            } else {
                // WebDocWf extension for save without references
                // The following line has been inserted
                if (references) {
                    // end of WebDocWf extension for save without references
                    // commit referenced DOs.
                    org.enhydra.barracuda.discRack.data.person.PersonDO owner_DO = getOwner();
                    if ( null != owner_DO ) {
                        if ( owner_DO.isLoaded() ) {
                            okToCommitOwner( owner_DO );
                            owner_DO.commit( dbt );
                        } else {
                            // since the referenced DO is not loaded,
                            // it cannot be dirty, so there is no need to commit it.
                        }
                    } else {
                        if ( ! false )
                            throw new RefAssertionException( "Cannot commit DiscDO ( " + toString() +
                                                             " ) because owner is not allowed to be null." );
                    }
 
                    // WebDocWf extension for save without references
                    // The following line has been inserted
                }
                // end of WebDocWf extension for save without references
            }
            if ( false ) {
                // This throw is here to keep the compiler happy
                // in the case of a DO that does not refer to other DOs.
                // In that case, the above delete/commit code blocks will be empty
                // and throw nothing.
                throw new DataObjectException( "foo" );   
            }
            if ( delete ) {
              dbt.delete( this );
            } else {
                if ( isLoaded() )
                    dbt.insert( this );   // dbt.insert() handles insertions and updates
            }
            if (ownTransaction) {
                dbt.commit(); // commit the transaction
            }
        } catch (SQLException sqle) {
            StringBuffer message = new StringBuffer("Failed to insert/update DO: ");
            message.append(sqle.getMessage());
            // rollback, if necessary
            if (ownTransaction) {
                try {
                    dbt.rollback();
                } catch (SQLException sqle2) {
                    message.insert(0,"\n");
                    message.insert(0,sqle2.getMessage());
                    message.insert(0,"Rollback failed: ");
                }
            }
            throw new SQLException(message.toString());
        } finally {
            // release the transaction, if any
            if (ownTransaction) {
                dbt.release();
            }
        }
    }
}
