
/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * org.enhydra.barracuda.discRack.data.person/PersonQuery.java
 *-----------------------------------------------------------------------------
 */

package org.enhydra.barracuda.discRack.data.person;

import org.enhydra.dods.DODS;
import com.lutris.dods.builder.generator.query.*;
import com.lutris.dods.builder.generator.dataobject.GenericDO;
import com.lutris.appserver.server.sql.*;
import org.enhydra.dods.cache.LRUCache;
import org.enhydra.dods.cache.QueryCacheItem;
import org.enhydra.dods.cache.QueryCacheItemImpl;
import org.enhydra.dods.cache.QueryResult;
import org.enhydra.dods.cache.DOShell;
import org.enhydra.dods.cache.Condition;
import org.enhydra.dods.statistics.Statistics;
import org.enhydra.dods.statistics.TableStatistics;
import org.enhydra.dods.statistics.QueryCacheStatistics;

import java.sql.*;
import java.util.*;
import java.math.*;
import java.util.Date;  // when I say Date, I don't mean java.sql.Date

// WebDocWf extension for DODS row instance based security
// The following line has been added
import org.webdocwf.dods.access.*;
// end of WebDocWf extension for DODS row instance based security

/**
 * PersonQuery is used to query the person table in the database.<BR>
 * It returns objects of type PersonDO.
 * <P>
 * General usage:
 * <P>
 *     In DODS:
 *        Create a Data Object named "Dog",
 *        and create an Attribute named "Name",
 *        and set that Attribute to "Can be queried."
 *        DODS will then generate the method DogQuery.setQueryName().
 * <P>
 *     In your Business Layer, prepare the query:<BR>
 * <P><PRE>
 *             DogQuery dq = new DogQuery();
 *             dq.setQueryName("Rex")
 *             if ( Rex is a reserved name )
 *                 dq.requireUniqueInstance();
 * </PRE>
 *     Then, get the query results one of two ways:
 * <P>
 *         #1:<PRE>
 *             String names = "";
 *             DogDO[] dogs = dq.getDOArray();
 *             for ( int i = 0; i < dogs.length; i++ ) {
 *                 names += dogs[i].getName() + " ";
 *             }
 * </PRE>
 *      or #2:<PRE>
 *             String names = "";
 *             DogDO dog;
 *             while ( null != ( dog = dq.getNextDO() ) ) {
 *                 names += dog.getName() + " ";
 *             }
 * </PRE>
 * <P>
 *     Note:   If <CODE>requireUniqueInstance()</CODE> was called,
 *             then <CODE>getDOArray()</CODE> or <CODE>getNextDO()</CODE>
 *             will throw an exception if more than one "Rex" was found.
 * <P>
 *     Note:   Results of the query will come from the Data Object cache if:
 *             -  The cache is available.
 *             -  Matches were found in the cache.
 *             -  No other tables (Data Objects of other types) were involved
 *                in the query.
 *                This can happen if you extend the <CODE>DogQuery</CODE> class
 *                and you make calls to the <CODE>QueryBuilder</CODE> object
 *                to add SQL involving other tables.
 *             If any of these conditions is not true,
 *             then any results from the query will come from the database.
 * <P>
 *     To reuse the query object, call:
 * <P><PRE>
 *             dq.reset();
 * </PRE>
 * @author NN
 * @version $Revision: 1.8 $
 */
public class PersonQuery implements Query {

    private QueryBuilder builder;

    /**
     * logical name of the database for which PersonQuery object has been created
     */
    private String logicalDatabase;

	private ResultSet resultSet = null;
	private boolean uniqueInstance  = false;
	private PersonDO[] DOs = null;
	private int arrayIndex = -1;
	private boolean needToRun = true;
	private LinkedHashMap cacheHits = null;
	private boolean isQueryByOId = false;
	private boolean hitDb = false;
    private boolean userHitDb = false;

	private QueryCacheItem queryItem = null;
	private String currentHandle = null;

    /**
     * Public constructor.
     */
    public PersonQuery() {
        builder = new QueryBuilder( "person", "person.*" );
        String dbName = PersonDO.getLogicalDBName();
        try {
            String vendor = DODS.getDatabaseManager().logicalDatabaseType(dbName);
            if (vendor != null) {
                builder.setDatabaseVendor(vendor);
                logicalDatabase = dbName;
            } else {
                builder.setDatabaseVendor();
                logicalDatabase =  DODS.getDatabaseManager().getDefaultDB();
            }
        } catch (Exception e) {
            builder.setDatabaseVendor();
            logicalDatabase =  DODS.getDatabaseManager().getDefaultDB();
        }
        builder.setUserStringAppendWildcard( false );
        builder.setUserStringTrim( false );
        reset();
    }



    /**
     * Return logical name of the database that PersonQuery object uses
     *
     * @return param logical database name
     *
     */
    public String getLogicalDatabase() {
        return logicalDatabase;
    }

    /**
     * Change logical database to another logical database (which name is dbName)
     *
     * @param dbName the logical name of the database
     *
     */
    public void setLogicalDatabase(String dbName) throws SQLException, DatabaseManagerException {
        // Support for DB independence added by Tanja Jovanovic 2002.12.03.
        String vendor = DODS.getDatabaseManager().logicalDatabaseType(dbName);
        if (vendor != null) {
            builder.setDatabaseVendor(vendor);
            logicalDatabase = dbName;
        } else {
            builder.setDatabaseVendor();
            logicalDatabase =  DODS.getDatabaseManager().getDefaultDB();
        }
        reset();
        // end of support for DB independence
    }

    //  WebDocWf extension for unique query results without SQL DISTINCT
    //  The following lines has been added:
    private boolean unique = false;

    /**
     * Set the unique flag of the query
     *
     * @param newUnique The unique flag for the query
     *
     * WebDocWf extension
     *
     */
    public void setUnique(boolean newUnique) { unique = newUnique; }

    /**
     * Get the unique flag of the query
     *
     * @return The unique flag of the query
     *
     * WebDocWf extension
     *
     */
    public boolean getUnique() { return unique; }
    //  end of WebDocWf extension for unique query results without SQL DISTINCT

    //  WebDocWf extension for skipping the n first rows of the result
    //  The following lines has been added:
    private int readSkip = 0;

    /**
     * Set the readSkip number of the query
     *
     * @param newreadSkip The number of results to skip
     *
     * WebDocWf extension
     *
     */
    public void setReadSkip(int newReadSkip) { readSkip = newReadSkip; }

    /**
     * Get the readSkip number of the query
     *
     * @return The number of rows which are skipped
     *
     * WebDocWf extension
     *
     */
    public int getReadSkip() { return readSkip; }
    // end of WebDocWf extension for skipping the n first rows of the result

    // WebDocWf extension for select rowcount limit
    // The following lines has been added:
    private int databaseLimit = 0;

    /**
     * Set the database limit of the query
     *
     * @param newLimit The limit for the query
     *
     * WebDocWf extension
     *
     */
    public void setDatabaseLimit(int newLimit) { databaseLimit = newLimit; }

    /**
     * Get the database limit of the query
     *
     * @return The database limit of the query
     *
     * WebDocWf extension
     *
     */
    public int getDatabaseLimit() { return databaseLimit; }

    private boolean databaseLimitExceeded = false;

    /**
     * Get the database limit exceeded flag of the query.
     *
     * @return The database limit exceeded flag of the query
     *   True if there would have been more rows than the limit, otherwise false.
     *
     * WebDocWf extension
     */
    public boolean getDatabaseLimitExceeded() { return databaseLimitExceeded; }
    // end of WebDocWf extension for select rowcount limit

    /**
     * Set that all queries go to database, not to cache.
     */
    public void hitDatabase() { userHitDb = true; }

    // WebDocWf extension for extended wildcard support
    // The following rows have been added:

    /**
     * Set user string wildcard.
     *
     * @param newUserStringWildcard New user string wildcard.
     *
     * WebDocWf extension
     */
    public void setUserStringWildcard(String newUserStringWildcard) {
        builder.setUserStringWildcard( newUserStringWildcard );
    }

    /**
     * Set user string single wildcard.
     *
     * @param newUserStringSingleWildcard New user string single wildcard.
     *
     * WebDocWf extension
     */
    public void setUserStringSingleWildcard(String newUserStringSingleWildcard) {
        builder.setUserStringSingleWildcard( newUserStringSingleWildcard );
    }

    /**
     * Set user string single wildcard escape.
     *
     * @param newUserStringSingleWildcardEscape New user string single wildcard escape.
     *
     * WebDocWf extension
     */
    public void setUserStringSingleWildcardEscape(String newUserStringSingleWildcardEscape) {
        builder.setUserStringSingleWildcardEscape( newUserStringSingleWildcardEscape );
    }

    /**
     * Set user string wildcard escape.
     *
     * @param newUserStringWildcardEscape New user string wildcard escape.
     *
     * WebDocWf extension
     */
    public void setUserStringWildcardEscape(String newUserStringWildcardEscape) {
        builder.setUserStringWildcardEscape( newUserStringWildcardEscape );
    }

    /**
     * Set user string append wildcard.
     *
     * @param userStringAppendWildcard New user string append wildcard.
     *
     * WebDocWf extension
     */
    public void setUserStringAppendWildcard(boolean userStringAppendWildcard ) {
        builder.setUserStringAppendWildcard( userStringAppendWildcard );
    }

     /**
     * Set user string trim.
     *
     * @param userStringTrim New user string trim.
     *
     * WebDocWf extension
     */
    public void setUserStringTrim(boolean userStringTrim ) {
        builder.setUserStringTrim( userStringTrim );
    }

    /**
     * Get user string wildcard.
     *
     * @return User string wildcard.
     *
     * WebDocWf extension
     */
    public String getUserStringWildcard() {
        return builder.getUserStringWildcard();
    }

    /**
     * Get user string single wildcard.
     *
     * @return User string single wildcard.
     *
     * WebDocWf extension
     */
    public String getUserStringSingleWildcard() {
        return builder.getUserStringSingleWildcard();
    }

    /**
     * Get user string single wildcard escape.
     *
     * @return User string single wildcard escape.
     *
     * WebDocWf extension
     */
    public String getUserStringSingleWildcardEscape() {
        return builder.getUserStringSingleWildcardEscape();
    }

    /**
     * Get user string wildcard escape.
     *
     * @return User string wildcard escape.
     *
     * WebDocWf extension
     */
    public String getUserStringWildcardEscape() {
        return builder.getUserStringWildcardEscape();
    }

    /**
     * Get user string append wildcard.
     *
     * @return User string append wildcard.
     *
     * WebDocWf extension
     */
    public boolean getUserStringAppendWildcard() {
        return builder.getUserStringAppendWildcard();
    }

    /**
     * Get user string trim.
     *
     * @return User string trim.
     *
     * WebDocWf extension
     */
    public boolean getUserStringTrim() {
        return builder.getUserStringTrim();
    }
    // end of WebDocWf extension for extended wildcard support

	/**
	 * Perform the query on the database, and prepare the
	 * array of returned DO objects.
	 *
	 * @exception DataObjectException If a database access error occurs.
	 * @exception NonUniqueQueryException If too many rows were found.
	 */
	private void getQueryByOIds(Vector DOs)
	throws DataObjectException {
		if (DOs.size() == 0)
			return;
		PersonDO DO = null;
		DOShell shell = null;
    	Date startQueryTime = new Date();
    	long queryTime = 0;
		for (int i=0; i<DOs.size(); i++) {
			shell = (DOShell)DOs.elementAt(i);
			DO = PersonDO.createExisting(shell.handle);
			shell.dataObject = DO;
        }
		Date stopQueryTime = new Date();
        queryTime = stopQueryTime.getTime() - startQueryTime.getTime();
	    PersonDO.statistics.updateQueryByOIdAverageTime((new Long(queryTime)).intValue(),DOs.size());
	}



	/**
	 * Perform the query on the database, and prepare the
	 * array of returned DO objects.
	 *
	 * @exception DataObjectException If a database access error occurs.
	 * @exception NonUniqueQueryException If too many rows were found.
	 */
	private void runQuery()
	throws DataObjectException, NonUniqueQueryException {
		needToRun = false;
		arrayIndex = -1;
		DBQuery dbQuery = null;
		Date startQueryTime = new Date();
		long queryTime = 0;
		boolean readDOs = false;
		try {
			QueryResult results = null;
			DOShell shell = null;
            if (PersonDO.cache != null) {
    			if (isQueryByOId) { // query by OId
    				results = new QueryResult();
    				if (currentHandle != null) { // select has one query by OId
   					    PersonDO DO = (PersonDO)PersonDO.cache.getDOByHandle(currentHandle);
    					if (DO != null) { // DO is found in the cache
    						if (queryItem.checkConditions(DO)){
    						    shell = new DOShell(DO);
    							results.DOs.add(shell);
    						}
                            ((QueryCacheStatistics)PersonDO.statistics).incrementCacheByOIdHitsNum();
    					} else { // DO isn't found in the cache
    						if (!PersonDO.cache.isFull())
    							hitDb = true;
    					}
    				} else { // select has more than one query by OId, there is no result for this query
    					hitDb = false;
    				}
    			} else { // other queries
    				String queryID = builder.getSQLwithParms()+"rs="+readSkip +"dl="+databaseLimit; // unique representation of query
    				if (hitDb) {
    					results = PersonDO.cache.getComplexQueryResults (logicalDatabase, queryID);
	                    ((TableStatistics)PersonDO.statistics).incrementComplexQueryNum();
    			    } else {
    					results = PersonDO.cache.getSimpleQueryResults (logicalDatabase, queryID);
	                    ((TableStatistics)PersonDO.statistics).incrementSimpleQueryNum();
                    }
    				if (results == null) { // query doesn't exist
    					queryItem.setQueryId(queryID);
    					if (hitDb)
    						PersonDO.cache.addComplexQuery(queryItem); // register complex query
    					else
    						PersonDO.cache.addSimpleQuery(queryItem); // register simple query
                        if (PersonDO.cache.isFull()) {
            				results = new QueryResult();
                        } else
    					    hitDb = true;
    				} else { // query found
        				if (hitDb)
    	                    ((QueryCacheStatistics)PersonDO.statistics).incrementComplexCacheHitsNum();
        			    else
    	                    ((QueryCacheStatistics)PersonDO.statistics).incrementSimpleCacheHitsNum();
                        int lazyTime = ((QueryCacheStatistics)PersonDO.statistics).getQueryByOIdAverageTime()*results.lazy.size();
                        if (lazyTime > queryItem.getTime())
        					hitDb = true;
        				else
    					    hitDb = false;
    				}
    			}
            } else
                hitDb = true;
			if ( hitDb || userHitDb) { // go to database

					dbQuery = PersonDO.createQuery();
				dbQuery.query( this );    // invokes executeQuery
				results = new QueryResult();
				// WebDocWf extension
				// The following lines have been added:
				int tempReadSkip=0;
				boolean bHasMoreResults = false;
				if (! unique) {
					while ( (bHasMoreResults = resultSet.next()) && (databaseLimit==0 || (results.DOs.size()<databaseLimit) ) ) {
						PersonDO newDO;

							newDO = PersonDO.createExisting ( resultSet );
      								if (tempReadSkip >= readSkip) {
      									newDO = PersonDO.addToCache(newDO);
      									queryItem.add(newDO);
      									shell = new DOShell(newDO);
      									results.DOs.add(shell);
      								} else {
      									tempReadSkip = tempReadSkip + 1;
      								}
					} // while
				} else { // if (! unique)
					HashSet hsResult = new HashSet(databaseLimit * 2);
					while ( (bHasMoreResults = resultSet.next()) && (databaseLimit==0 || (hsResult.size()<databaseLimit) ) ) {
						PersonDO newDO;
						    newDO = PersonDO.createExisting ( resultSet );
								if (tempReadSkip >= readSkip) {
									if (!hsResult.contains(newDO.getHandle())) {
										newDO = PersonDO.addToCache(newDO);
										queryItem.add(newDO);
										shell = new DOShell(newDO);
      									results.DOs.add(shell);
										hsResult.add(newDO.getHandle());
									} else {
										tempReadSkip = tempReadSkip +1;
									}
								} else {
									if (!hsResult.contains(newDO.getHandle()))
										tempReadSkip = tempReadSkip +1;
								}
                    }
				} // else (! unique)
				if ((results.DOs.size()==databaseLimit)&& bHasMoreResults) {
					resultSet.close();
					databaseLimitExceeded = true;
				}
        		Date stopQueryTime = new Date();
                queryTime = stopQueryTime.getTime() - startQueryTime.getTime();
                queryItem.setTime((new Long(queryTime)).intValue());
			} else { // if ( hitDb )
				if (PersonDO.cache != null) {
//    				System.out.println("Pre Person"+results.toString(true));
    				if (PersonDO.cache.isFull()) {
    					if (readSkip<cacheHits.size()) {
    						Vector vect = new Vector(cacheHits.values());
    						results.DOs = new Vector(vect.subList(readSkip, vect.size()));
                            readDOs = true;
    						// executeQuery saw cache hits
    					} else {
    						results.DOs = new Vector();
    					}
    				} else {
    					getQueryByOIds(results.lazy);
    				}
//    				System.out.println("Posle Person"+results.toString(true));
     				((QueryCacheStatistics)PersonDO.statistics).incrementCacheHitsNum();
//     				((QueryCacheStatistics)PersonDO.getStatistics()).show();
                }
			}
			// end of WebDocWf extension
			if ( results.DOs.size() > 1 && uniqueInstance )
				throw new NonUniqueQueryException("Too many rows returned from database" );
			DOs = new PersonDO [ results.DOs.size() ];
            if (readDOs)
    			for ( int i = 0; i < results.DOs.size(); i++ )
    				DOs[i] = (PersonDO)results.DOs.elementAt(i);
            else
    			for ( int i = 0; i < results.DOs.size(); i++ )
    				DOs[i] = (PersonDO)((DOShell)results.DOs.elementAt(i)).dataObject;
		    PersonDO.statistics.incrementQueryNum();
			arrayIndex = 0;
            if (!isQueryByOId)
			    PersonDO.statistics.updateQueryAverageTime((new Long(queryTime)).intValue());
			else
                PersonDO.statistics.incrementQueryByOIdNum();
		} catch ( SQLException se ) {
			if (null == se.getSQLState() ) {
				throw new DataObjectException("Unknown SQLException", se );
			}
			if ( se.getSQLState().startsWith("02") && se.getErrorCode() == 100 ) {
				throw new DataObjectException("Update or delete DO is out of synch", se );
			} else if (     se.getSQLState().equals("S1000") &&     se.getErrorCode() == -268) {
				throw new DataObjectException("Integrity constraint violation", se );
			} else {
				throw new DataObjectException( "Data Object Error", se );
			}
		} catch ( ObjectIdException oe ) {
			throw new DataObjectException( "Object ID Error", oe );
		} catch ( DatabaseManagerException de ) {
			throw new DataObjectException( "Database connection Error", de );
		} finally {
			if ( null != dbQuery )
				dbQuery.release();
		}
	}

    /**
     * Limit the number of rows (DOs) returned.
     * NOTE: When setting a limit on rows returned by a query,
     * you usually want to use a call to an addOrderBy method
     * to cause the most interesting rows to be returned first.

     *
     * @param maxRows Max number of rows (DOs) returned.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception NonUniqueQueryException If too many rows were found.
     */
    public void setMaxRows( int maxRows )
    throws DataObjectException, NonUniqueQueryException {
        userHitDb = true;
        builder.setMaxRows( maxRows );
    }

    /**
     * Return array of DOs constructed from ResultSet returned by query.
     *
     * @return Array of DOs constructed from ResultSet returned by query.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception NonUniqueQueryException If too many rows were found.
     */
    public PersonDO[] getDOArray()
    throws DataObjectException, NonUniqueQueryException {
        if ( needToRun )
            runQuery();
        return DOs;
    }

    /**
     * Return successive DOs from array built from ResultSet returned by query.
     *
     * @return DOs from array built from ResultSet returned by query.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception NonUniqueQueryException If too many rows were found.
     */
    public PersonDO getNextDO()
    throws DataObjectException, NonUniqueQueryException {
        if ( needToRun )
            runQuery();
        if ( null == DOs ) {
            /* This should never happen.
             * runQuery() should either throw an exception
             * or create an array of DOs, possibly of zero length.
             */
            return null;
        }
        if ( arrayIndex < DOs.length )
            return DOs[ arrayIndex++ ];
        return null;
    }


    /**
     * Set the OID to query.
     * WARNING!  This method assumes that table <CODE>person</CODE>
     * has a column named <CODE>"oid"</CODE>.
     * This method is called from the DO classes to retrieve an object by id.
     *
     * @param oid The object id to query.
     */
    public void setQueryOId(ObjectId oid) {
        // Remove from cacheHits any DO that does not meet this
        // setQuery requirement.
		String handle = oid.toString();

        if (PersonDO.cache != null && PersonDO.cache.isFull()) {
			isQueryByOId = true;
            PersonDO DO = (PersonDO)cacheHits.get(handle);
            if (DO != null) {
                cacheHits = new LinkedHashMap();
                cacheHits.put(handle,DO);
            }
        } else {
			if (isQueryByOId) { // query by OId already has been invoked
				if (!currentHandle.equals(handle)) // two different OIds are invoked
					currentHandle = null;
			} else {
				currentHandle =  handle;
				isQueryByOId = true;
			}
        }
        // Also prepare the SQL needed to query the database
        // in case there is no cache, or the query involves other tables.
        builder.addWhere( PersonDO.PrimaryKey, oid.toBigDecimal(), QueryBuilder.EQUAL );
        PersonDO.statistics.incrementQueryByOIdNum();
    }

    /**
     * Set the object handle to query.
     * This is a variant of setQueryOId().
     *
     * @param handle The string version of the id to query.
     */
    public void setQueryHandle(String handle)
    throws ObjectIdException {
        ObjectId oid = new ObjectId( handle );
        setQueryOId( oid );
    }


    /**
     * Set "unique instance" assertion bit.
     * The first call to the next() method will throw an exception
     * if more than one object was found.
     */
    public void requireUniqueInstance() {
        uniqueInstance = true;
    }

    /**
     * Reset the query parameters.
     */
    public void reset() {
        if (PersonDO.cache != null && PersonDO.cache.isFull()) {
            cacheHits = new LinkedHashMap(PersonDO.cache.getCacheContent());
        }
        DOs = null;
        uniqueInstance  = false;
        needToRun       = true;
        builder.reset();
        if (PersonDO.cache != null)
            queryItem = PersonDO.cache.newQueryCacheItemInstance(logicalDatabase);
        else
            queryItem = new QueryCacheItemImpl(logicalDatabase);
        userHitDb = false;
    }

    /**
     * Return the appropriate QueryBuilder flag for selecting
     * exact matches (SQL '=') or inexact matches (SQL 'matches').
     *
     * @return boolean True if it is exact match, otherwise false.
     *
     */
    private boolean exactFlag( boolean exact ) {
        return exact ? QueryBuilder.EXACT_MATCH : QueryBuilder.NOT_EXACT;
    }

    //
    // Implementation of Query interface
    //

    /**
     * Method to query objects from the database.
     * The following call in runQuery()
     *          dbQuery.query( this );
     * causes the dbQuery object to invoke
     *      executeQuery()
     *
     * @param conn Handle to database connection.
     *
     * @return ResultSet with the results of the query.
     *
     * @exception java.sql.SQLException If a database access error occurs.
     */
    public ResultSet executeQuery(DBConnection conn)
    throws SQLException     {
        resultSet = builder.executeQuery( conn );
        return resultSet;
    }

    /**
     * WARNING!  This method is disabled.
     * It's implementation is forced by the Query interface.
     * This method is disabled for 2 reasons:
     * 1)  the getDOArray() and getNextDO() methods are better
     *     because they return DOs instead of JDBC objects.
     * 2)  the createExisting() method throws an exception
     *     that we cannot reasonably handle here,
     *     and that we cannot throw from here.
     *
     * @param rs JDBC result set from which the next object
     *   will be instantiated.
     *
     * @exception java.sql.SQLException
     *   If a database access error occurs.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an invalid object id was queried from the database.
     */
    public Object next(ResultSet rs)
    throws SQLException, ObjectIdException {
        // TODO: It would be nice to throw an unchecked exception here
        // (an exception that extends RuntimeException)
        // that would be guaranteed to appear during application testing.
        throw new ObjectIdException("next() should not be used.  Use getNextDO() instead." );
        //return null;
    }


    // WebDocWf extension for extended wildcard support
    // The following lines have been added:
    /**
     * Convert a String with user wildcards into a string with DB wildcards
     *
     * @param userSearchValue The string with user wildcards
     *
     * @return The string with DB wildcards
     *
     * WebDocWf extension
     *
     */
    public String convertUserSearchValue( String userSearchValue ) {
        return builder.convertUserSearchValue( userSearchValue );
    }

    /**
     * Check whether a string contains DB wildcards
     *
     * @param dbSearchValue The string with possible DB wildcards
     *
     * @return Whether a string contains DB wildcards
     *
     * WebDocWf extension
     *
     */
    public boolean containsWildcards( String dbSearchValue ) {
        return builder.containsWildcards( dbSearchValue );
    }
    // end of WebDocWf extension for extended wildcard support

    // WebDocWf extension for query row counter
    // The following lines have been added:
    /**
     * Get the rowcount of the query
     * If possible, do it without reading all rows
     *
     * @return The row count
     *
     * WebDocWf extension
     *
     */
    public int getCount()
    throws NonUniqueQueryException, DataObjectException, SQLException, DatabaseManagerException     {
        int rowCount=0;
        if (needToRun && databaseLimit==0) {
            rowCount = selectCount();
        } else {
            if (needToRun) runQuery();
                rowCount = DOs.length;
        }
        return rowCount;
    }

    /**
     * Get the rowcount of the query by using count(*) in the DB
     *
     * @return The row count
     *
     * WebDocWf extension
     *
     */
    public int selectCount()
    throws SQLException, DatabaseManagerException {
        int rowCount=0;
        String tempClause = builder.getSelectClause();
        builder.setSelectClause(" count(*) as \"counter\" ");
        DBQuery dbQuery;

            dbQuery = PersonDO.createQuery();
        dbQuery.query(this);
        resultSet.next();
        rowCount=resultSet.getInt("counter");
        resultSet.close();
        dbQuery.release();
        builder.close();
        resultSet = null;
        builder.setSelectClause(tempClause);
        return rowCount;
    }
    // end of WebDocWf extension for query row counter

    /**
     * Set the login to query, with a QueryBuilder comparison operator.
     *
     * @param x The login of the person to query.
     * @param cmp_op QueryBuilder comparison operator to use.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryLogin(String x, String cmp_op )
    throws DataObjectException, QueryException {
        // WebDocWf extension for extended wildcard support
        // The following lines have been added:

        Condition cond = new Condition(PersonDO.COLUMN_LOGIN,x,cmp_op);
        queryItem.addCond(cond);
        if (cmp_op.equals(QueryBuilder.CASE_INSENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.CASE_SENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.USER_CASE_SENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.USER_CASE_INSENSITIVE_MATCH)) {
            hitDb = true;
        } else {
            // end of WebDocWf extension for extended wildcard support


            if (PersonDO.cache != null && PersonDO.cache.isFull()) {
                // Remove from cacheHits any DOs that do not meet this
                // setQuery requirement.
                for ( Iterator iter = (new HashSet(cacheHits.values())).iterator(); iter.hasNext();) {
                    PersonDO DO = (PersonDO)iter.next();
                    String m = DO.getLogin();
                    if(!QueryBuilder.compare( m, x, cmp_op )) {
                        try {
                            String handle = DO.getHandle();
                            cacheHits.remove(handle);
                        } catch (DatabaseManagerException e) {
                            throw new DataObjectException("Error in loading data object's handle.");
                        }
                    }
                }
            }
        }

        // Also prepares the SQL needed to query the database
        // in case there is no cache, or the query involves other tables.
        // WebDocWf patch for correct queries in fully cached objects
        // The following line has been put under comments:
        // if ( partOrLru || hitDb )
        // end of WebDocWf patch for correct queries in fully cached objects

        builder.addWhere( PersonDO.Login, x, cmp_op );
    }

    /**
     * Set the login to query, with a QueryBuilder comparison operator.
     *
     * @param x The login of the person to query.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryLogin(String x)
    throws DataObjectException, QueryException {
         setQueryLogin(x, QueryBuilder.EQUAL);
    }


    /**
     * Add login to the ORDER BY clause.
     * NOTE: The DO cache does not yet support the Order By operation.
     * Using the addOrderBy method forces the query to hit the database.
     *
     * @param direction_flag  True for ascending order, false for descending
     */
    public void addOrderByLogin(boolean direction_flag) {
        hitDb = true;
        builder.addOrderByColumn("login", (direction_flag) ? "ASC" : "DESC");
    }


    /**
     * Add login to the ORDER BY clause.  This convenience
     * method assumes ascending order.
     * NOTE: The DO cache does not yet support the Order By operation.
     * Using the addOrderBy method forces the query to hit the database.
     */
    public void addOrderByLogin() {
        hitDb = true;
        builder.addOrderByColumn("login","ASC");
    }



    /**
     * Set the login to query with a user wildcard string
     *
     * @param x The login of the person to query with user wildcards
     *
     * @exception DataObjectException If a database access error occurs.
     *
     * @deprecated Use comparison operators instead
     *
     * WebDocWf extension
     *
     */
    public void setUserMatchLogin( String x )
    throws DataObjectException, QueryException {
        String y = convertUserSearchValue( x );
        setDBMatchLogin( y );
    }

    /**
     * Set the login to query with a DB wildcard string
     *
     * @param x The login of the person to query with DB wildcards
     *
     * @exception DataObjectException If a database access error occurs.
     *
     * @deprecated Use comparison operators instead
     *
     * WebDocWf extension
     *
     */
    public void setDBMatchLogin(String x )
    throws DataObjectException, QueryException {
        if (containsWildcards(x) || builder.getUserStringAppendWildcard()) {
            builder.addMatchClause( PersonDO.Login, x );
            hitDb = true;
        } else
            setQueryLogin( x, QueryBuilder.EQUAL );
    }


    // end of WebDocWf extension for extended wildcard support

    /**
     * Set the password to query, with a QueryBuilder comparison operator.
     *
     * @param x The password of the person to query.
     * @param cmp_op QueryBuilder comparison operator to use.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryPassword(String x, String cmp_op )
    throws DataObjectException, QueryException {
        // WebDocWf extension for extended wildcard support
        // The following lines have been added:

        Condition cond = new Condition(PersonDO.COLUMN_PASSWORD,x,cmp_op);
        queryItem.addCond(cond);
        if (cmp_op.equals(QueryBuilder.CASE_INSENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.CASE_SENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.USER_CASE_SENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.USER_CASE_INSENSITIVE_MATCH)) {
            hitDb = true;
        } else {
            // end of WebDocWf extension for extended wildcard support


            if (PersonDO.cache != null && PersonDO.cache.isFull()) {
                // Remove from cacheHits any DOs that do not meet this
                // setQuery requirement.
                for ( Iterator iter = (new HashSet(cacheHits.values())).iterator(); iter.hasNext();) {
                    PersonDO DO = (PersonDO)iter.next();
                    String m = DO.getPassword();
                    if(!QueryBuilder.compare( m, x, cmp_op )) {
                        try {
                            String handle = DO.getHandle();
                            cacheHits.remove(handle);
                        } catch (DatabaseManagerException e) {
                            throw new DataObjectException("Error in loading data object's handle.");
                        }
                    }
                }
            }
        }

        // Also prepares the SQL needed to query the database
        // in case there is no cache, or the query involves other tables.
        // WebDocWf patch for correct queries in fully cached objects
        // The following line has been put under comments:
        // if ( partOrLru || hitDb )
        // end of WebDocWf patch for correct queries in fully cached objects

        builder.addWhere( PersonDO.Password, x, cmp_op );
    }

    /**
     * Set the password to query, with a QueryBuilder comparison operator.
     *
     * @param x The password of the person to query.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryPassword(String x)
    throws DataObjectException, QueryException {
         setQueryPassword(x, QueryBuilder.EQUAL);
    }


    /**
     * Add password to the ORDER BY clause.
     * NOTE: The DO cache does not yet support the Order By operation.
     * Using the addOrderBy method forces the query to hit the database.
     *
     * @param direction_flag  True for ascending order, false for descending
     */
    public void addOrderByPassword(boolean direction_flag) {
        hitDb = true;
        builder.addOrderByColumn("password", (direction_flag) ? "ASC" : "DESC");
    }


    /**
     * Add password to the ORDER BY clause.  This convenience
     * method assumes ascending order.
     * NOTE: The DO cache does not yet support the Order By operation.
     * Using the addOrderBy method forces the query to hit the database.
     */
    public void addOrderByPassword() {
        hitDb = true;
        builder.addOrderByColumn("password","ASC");
    }



    /**
     * Set the password to query with a user wildcard string
     *
     * @param x The password of the person to query with user wildcards
     *
     * @exception DataObjectException If a database access error occurs.
     *
     * @deprecated Use comparison operators instead
     *
     * WebDocWf extension
     *
     */
    public void setUserMatchPassword( String x )
    throws DataObjectException, QueryException {
        String y = convertUserSearchValue( x );
        setDBMatchPassword( y );
    }

    /**
     * Set the password to query with a DB wildcard string
     *
     * @param x The password of the person to query with DB wildcards
     *
     * @exception DataObjectException If a database access error occurs.
     *
     * @deprecated Use comparison operators instead
     *
     * WebDocWf extension
     *
     */
    public void setDBMatchPassword(String x )
    throws DataObjectException, QueryException {
        if (containsWildcards(x) || builder.getUserStringAppendWildcard()) {
            builder.addMatchClause( PersonDO.Password, x );
            hitDb = true;
        } else
            setQueryPassword( x, QueryBuilder.EQUAL );
    }


    // end of WebDocWf extension for extended wildcard support

    /**
     * Set the firstname to query, with a QueryBuilder comparison operator.
     *
     * @param x The firstname of the person to query.
     * @param cmp_op QueryBuilder comparison operator to use.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryFirstname(String x, String cmp_op )
    throws DataObjectException, QueryException {
        // WebDocWf extension for extended wildcard support
        // The following lines have been added:

        Condition cond = new Condition(PersonDO.COLUMN_FIRSTNAME,x,cmp_op);
        queryItem.addCond(cond);
        if (cmp_op.equals(QueryBuilder.CASE_INSENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.CASE_SENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.USER_CASE_SENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.USER_CASE_INSENSITIVE_MATCH)) {
            hitDb = true;
        } else {
            // end of WebDocWf extension for extended wildcard support


            if (PersonDO.cache != null && PersonDO.cache.isFull()) {
                // Remove from cacheHits any DOs that do not meet this
                // setQuery requirement.
                for ( Iterator iter = (new HashSet(cacheHits.values())).iterator(); iter.hasNext();) {
                    PersonDO DO = (PersonDO)iter.next();
                    String m = DO.getFirstname();
                    if(!QueryBuilder.compare( m, x, cmp_op )) {
                        try {
                            String handle = DO.getHandle();
                            cacheHits.remove(handle);
                        } catch (DatabaseManagerException e) {
                            throw new DataObjectException("Error in loading data object's handle.");
                        }
                    }
                }
            }
        }

        // Also prepares the SQL needed to query the database
        // in case there is no cache, or the query involves other tables.
        // WebDocWf patch for correct queries in fully cached objects
        // The following line has been put under comments:
        // if ( partOrLru || hitDb )
        // end of WebDocWf patch for correct queries in fully cached objects

        builder.addWhere( PersonDO.Firstname, x, cmp_op );
    }

    /**
     * Set the firstname to query, with a QueryBuilder comparison operator.
     *
     * @param x The firstname of the person to query.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryFirstname(String x)
    throws DataObjectException, QueryException {
         setQueryFirstname(x, QueryBuilder.EQUAL);
    }


    /**
     * Add firstname to the ORDER BY clause.
     * NOTE: The DO cache does not yet support the Order By operation.
     * Using the addOrderBy method forces the query to hit the database.
     *
     * @param direction_flag  True for ascending order, false for descending
     */
    public void addOrderByFirstname(boolean direction_flag) {
        hitDb = true;
        builder.addOrderByColumn("firstname", (direction_flag) ? "ASC" : "DESC");
    }


    /**
     * Add firstname to the ORDER BY clause.  This convenience
     * method assumes ascending order.
     * NOTE: The DO cache does not yet support the Order By operation.
     * Using the addOrderBy method forces the query to hit the database.
     */
    public void addOrderByFirstname() {
        hitDb = true;
        builder.addOrderByColumn("firstname","ASC");
    }



    /**
     * Set the firstname to query with a user wildcard string
     *
     * @param x The firstname of the person to query with user wildcards
     *
     * @exception DataObjectException If a database access error occurs.
     *
     * @deprecated Use comparison operators instead
     *
     * WebDocWf extension
     *
     */
    public void setUserMatchFirstname( String x )
    throws DataObjectException, QueryException {
        String y = convertUserSearchValue( x );
        setDBMatchFirstname( y );
    }

    /**
     * Set the firstname to query with a DB wildcard string
     *
     * @param x The firstname of the person to query with DB wildcards
     *
     * @exception DataObjectException If a database access error occurs.
     *
     * @deprecated Use comparison operators instead
     *
     * WebDocWf extension
     *
     */
    public void setDBMatchFirstname(String x )
    throws DataObjectException, QueryException {
        if (containsWildcards(x) || builder.getUserStringAppendWildcard()) {
            builder.addMatchClause( PersonDO.Firstname, x );
            hitDb = true;
        } else
            setQueryFirstname( x, QueryBuilder.EQUAL );
    }


    // end of WebDocWf extension for extended wildcard support

    /**
     * Set the lastname to query, with a QueryBuilder comparison operator.
     *
     * @param x The lastname of the person to query.
     * @param cmp_op QueryBuilder comparison operator to use.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryLastname(String x, String cmp_op )
    throws DataObjectException, QueryException {
        // WebDocWf extension for extended wildcard support
        // The following lines have been added:

        Condition cond = new Condition(PersonDO.COLUMN_LASTNAME,x,cmp_op);
        queryItem.addCond(cond);
        if (cmp_op.equals(QueryBuilder.CASE_INSENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.CASE_SENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.USER_CASE_SENSITIVE_MATCH) || cmp_op.equals(QueryBuilder.USER_CASE_INSENSITIVE_MATCH)) {
            hitDb = true;
        } else {
            // end of WebDocWf extension for extended wildcard support


            if (PersonDO.cache != null && PersonDO.cache.isFull()) {
                // Remove from cacheHits any DOs that do not meet this
                // setQuery requirement.
                for ( Iterator iter = (new HashSet(cacheHits.values())).iterator(); iter.hasNext();) {
                    PersonDO DO = (PersonDO)iter.next();
                    String m = DO.getLastname();
                    if(!QueryBuilder.compare( m, x, cmp_op )) {
                        try {
                            String handle = DO.getHandle();
                            cacheHits.remove(handle);
                        } catch (DatabaseManagerException e) {
                            throw new DataObjectException("Error in loading data object's handle.");
                        }
                    }
                }
            }
        }

        // Also prepares the SQL needed to query the database
        // in case there is no cache, or the query involves other tables.
        // WebDocWf patch for correct queries in fully cached objects
        // The following line has been put under comments:
        // if ( partOrLru || hitDb )
        // end of WebDocWf patch for correct queries in fully cached objects

        builder.addWhere( PersonDO.Lastname, x, cmp_op );
    }

    /**
     * Set the lastname to query, with a QueryBuilder comparison operator.
     *
     * @param x The lastname of the person to query.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryLastname(String x)
    throws DataObjectException, QueryException {
         setQueryLastname(x, QueryBuilder.EQUAL);
    }


    /**
     * Add lastname to the ORDER BY clause.
     * NOTE: The DO cache does not yet support the Order By operation.
     * Using the addOrderBy method forces the query to hit the database.
     *
     * @param direction_flag  True for ascending order, false for descending
     */
    public void addOrderByLastname(boolean direction_flag) {
        hitDb = true;
        builder.addOrderByColumn("lastname", (direction_flag) ? "ASC" : "DESC");
    }


    /**
     * Add lastname to the ORDER BY clause.  This convenience
     * method assumes ascending order.
     * NOTE: The DO cache does not yet support the Order By operation.
     * Using the addOrderBy method forces the query to hit the database.
     */
    public void addOrderByLastname() {
        hitDb = true;
        builder.addOrderByColumn("lastname","ASC");
    }



    /**
     * Set the lastname to query with a user wildcard string
     *
     * @param x The lastname of the person to query with user wildcards
     *
     * @exception DataObjectException If a database access error occurs.
     *
     * @deprecated Use comparison operators instead
     *
     * WebDocWf extension
     *
     */
    public void setUserMatchLastname( String x )
    throws DataObjectException, QueryException {
        String y = convertUserSearchValue( x );
        setDBMatchLastname( y );
    }

    /**
     * Set the lastname to query with a DB wildcard string
     *
     * @param x The lastname of the person to query with DB wildcards
     *
     * @exception DataObjectException If a database access error occurs.
     *
     * @deprecated Use comparison operators instead
     *
     * WebDocWf extension
     *
     */
    public void setDBMatchLastname(String x )
    throws DataObjectException, QueryException {
        if (containsWildcards(x) || builder.getUserStringAppendWildcard()) {
            builder.addMatchClause( PersonDO.Lastname, x );
            hitDb = true;
        } else
            setQueryLastname( x, QueryBuilder.EQUAL );
    }


    // end of WebDocWf extension for extended wildcard support

    /**
     * Returns the <code>QueryBuilder</code> that this <code>PersonQuery</code>
     * uses to construct and execute database queries.
     * <code>PersonQuery.setQueryXXX</code> methods use 
     * the <code>QueryBuilder</code> to
     * append SQL expressions to the <code>"WHERE"</code> clause to be executed.
     * The <code>QueryBuilder.addEndClause method.</code> can be used to
     * append freeform SQL expressions to the <code>WHERE</code> clause,
     * e.g. "ORDER BY name".
     *
     * <b>Notes regarding cache-enabled DO classes:</b>
     * DO classes can be cache-enabled.  
     * If when using a <code>PersonQuery</code>, the application developer
     * <b>does not call</b> <code>getQueryBuilder</code>,
     * then <code>PersonQuery.setQueryXXX</code> methods 
     * simply prune the DO cache and return the remaining results.
     * However, a <code>QueryBuilder</code> builds
     * <CODE>SELECT</CODE> statements for execution by the actual database,
     * and never searches the built-in cache for the DO.
     * So, if the DO class is cache-enabled, and <code>getQueryBuilder</code>
     * is called, this <CODE>PersonQuery</CODE> object ignores the cache 
     * and hits the actual database.
     * 
     * @return QueryBuilder that is used to construct and execute database queries.   
     */
    public QueryBuilder getQueryBuilder() {
        hitDatabase();
        return builder;
    }

    /**
     * Insert an <CODE>OR</CODE> between <CODE>WHERE</CODE> clauses.
     * Example:  find all the persons named Bob or Robert:
     * <CODE>
     *    PersonQuery pq = new PersonQuery();
     *    pq.setQueryFirstName( "Bob" );
     *    pq.or();
     *    pq.setQueryFirstName( "Robert" );
     * </CODE>
     * 
     * Note:  Calls to <CODE>setQueryXxx</CODE> methods
     * are implicitly <CODE>AND</CODE>ed together,
     * so the following example will always return nothing:
     * <CODE>
     *    PersonQuery pq = new PersonQuery();
     *    pq.setQueryFirstName( "Bob" );
     *    // AND automatically inserted here.
     *    pq.setQueryFirstName( "Robert" );
     * </CODE>
    
     * 
     * @see QueryBuilder to construct more elaborate queries.
     * @author Jay Gunter
     */
    public void or() {
        hitDb = true;
        builder.addWhereOr();
    }

    /**
     * Place an open parenthesis in the <CODE>WHERE</CODE> clause.
     * Example usage:  find all the Bobs and Roberts who are 5 or 50 years old:
     * <CODE>
     *    PersonQuery pq = new PersonQuery();
     *    pq.openParen();
     *       pq.setQueryFirstName( "Bob" );
     *       pq.or();
     *       pq.setQueryFirstName( "Robert" );
     *    pq.closeParen();
     *    // AND automatically inserted here.
     *    pq.openParen();
     *       pq.setQueryAge( 5 );
     *       pq.or();
     *       pq.setQueryAge( 50 );
     *    pq.closeParen();
     * </CODE>
    
     * 
     * @see QueryBuilder to construct more elaborate queries.
     * @author Jay Gunter
     */
    public void openParen() {
        hitDb = true;
        builder.addWhereOpenParen();
    }

    /**
     * Place a closing parenthesis in the <CODE>WHERE</CODE> clause.
    
     * 
     * @see openParen
     * @author Jay Gunter
     */
    public void closeParen() {
        hitDb = true;
        builder.addWhereCloseParen();
    }
}
