/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: LoginScreen.java,v 1.1 2003/04/14 12:24:56 sinisa Exp $
 */
package org.enhydra.barracuda.discRack.pres.screens;

import java.io.*;
import java.util.*;
import java.net.*;
import javax.servlet.*;
import javax.servlet.http.*;

import org.w3c.dom.*;
import org.w3c.dom.html.*;

import com.lutris.appserver.server.Enhydra;
import com.lutris.logging.LogChannel;

import org.enhydra.xml.xmlc.*;

import org.enhydra.barracuda.core.comp.*;
import org.enhydra.barracuda.core.event.*;
import org.enhydra.barracuda.core.event.helper.*;
import org.enhydra.barracuda.core.forms.*;
import org.enhydra.barracuda.core.forms.validators.*;
import org.enhydra.barracuda.core.util.*;
import org.enhydra.barracuda.core.util.data.*;
import org.enhydra.barracuda.core.util.data.Collections;
import org.enhydra.barracuda.core.util.dom.*;
import org.enhydra.barracuda.core.util.http.*;
import org.apache.log4j.*;

import org.enhydra.barracuda.discRack.biz.*;
import org.enhydra.barracuda.discRack.biz.person.*;
import org.enhydra.barracuda.discRack.pres.events.*;
import org.enhydra.barracuda.discRack.pres.services.*;
import org.enhydra.barracuda.discRack.pres.xmlc.*;


/**
 * <p>Event handlers (both Controller and View) for the 
 * Login screen
 *
 * <p>We handle the following Control events:
 * <ul>
 *		<li>GetLogin - if we are already logged in, fire a DoLogin 
 *			event. Otherwise, just fire a RenderLogin event.</li>
 *		<li>DoLogin - create the login form, map the request to the 
 *			form and then validate it. If there weren't any errors, 
 *			fire a GetCatalog event. Otherwise, fire a GetLogin event
 *			(saving the form and any validation errors in the event 
 *			context so that the RenderLogin can update the screen 
 *			correctly.</li>
 *		<li>DoLogout - Fire a RenderLogout event</li>
 *		<li>ValidatedEvent - Whenever a ValidatedEvent is fired, 
 *			we must ensure the user is actually logged in. If so, 
 *			do nothing. If not, redirect the user to the login screen 
 *			by firing a GetLogin event.</li>
 * </ul>
 *
 * <p>We handle the following View events:
 * <ul>
 *		<li>RenderLogin - Generate the Login screen. If the event context
 *			contains a LoginForm, we repopulate the screen from the form.
 *			If the context contains login errs, we will show those as well</li>
 *		<li>RenderLogout - Generate the Logout screen</li>
 *		<li>SampleError - Generate an error screen</li>
 * </ul>
 *
 * <p>We define the following Forms:
 * <ul>
 *		<li>LoginForm - contains 2 elements: USER and PASSWORD, both of 
 *			which are Strings and take NotNullValidators. In addition, 
 *			login form uses LoginValidator, which validates the form by 
 *			making sure that the user record exists and the password entered
 *			matches the password for that user.</li>
 * </ul>
 *
 * <p>You can <a href="sm_barracudaDiscRack_login.gif">click here to see a diagram...</a>
 */
public class LoginScreen extends DefaultEventGateway {

	//public constants
//	public static int showDebug = 0;
	protected static Logger logger = Logger.getLogger(LoginScreen.class.getName());

	//private constants
	private static final String LOGIN_FORM = LoginScreen.class.getName()+".LoginForm";		//LoginForm
	private static final String LOGIN_ERR = LoginScreen.class.getName()+".LoginErr";		//ValidationException

	//this defines the various event handlers
	private ListenerFactory getLoginFactory = 			new DefaultListenerFactory() {public BaseEventListener getInstance() {return new GetLoginHandler();}			public String getListenerID() {return getID(GetLoginHandler.class);}};
	private ListenerFactory doLoginFactory = 			new DefaultListenerFactory() {public BaseEventListener getInstance() {return new DoLoginHandler();}				public String getListenerID() {return getID(DoLoginHandler.class);}};				
	private ListenerFactory doLogoutFactory = 			new DefaultListenerFactory() {public BaseEventListener getInstance() {return new DoLogoutHandler();}			public String getListenerID() {return getID(DoLogoutHandler.class);}};				
	private ListenerFactory validateLoginFactory = 		new DefaultListenerFactory() {public BaseEventListener getInstance() {return new ValidateLoginHandler();}		public String getListenerID() {return getID(ValidateLoginHandler.class);}};				
	private ListenerFactory renderLoginFactory = 		new DefaultListenerFactory() {public BaseEventListener getInstance() {return new RenderLoginHandler();}			public String getListenerID() {return getID(RenderLoginHandler.class);}};				
	private ListenerFactory renderLogoutFactory = 		new DefaultListenerFactory() {public BaseEventListener getInstance() {return new RenderLogoutHandler();}		public String getListenerID() {return getID(RenderLogoutHandler.class);}};				
	private ListenerFactory renderSampleErrorFactory = 	new DefaultListenerFactory() {public BaseEventListener getInstance() {return new RenderSampleErrorHandler();}	public String getListenerID() {return getID(RenderSampleErrorHandler.class);}};				

	//private vars
	
	
	//-------------------- DefaultEventGateway -------------------
	/**
	 * Public constructor
	 */
	public LoginScreen() {
		//specify who's interested in what
		specifyLocalEventInterests(getLoginFactory, GetLogin.class);
		specifyLocalEventInterests(doLoginFactory, DoLogin.class);
		specifyLocalEventInterests(doLogoutFactory, DoLogout.class);
		specifyLocalEventInterests(validateLoginFactory, ValidatedEvent.class);
		specifyLocalEventInterests(renderLoginFactory, RenderLogin.class);
		specifyLocalEventInterests(renderLogoutFactory, RenderLogout.class);
		specifyLocalEventInterests(renderSampleErrorFactory, SampleError.class);
	}


	//------------------------------------------------------------
	//             Model 2 - Controller Event Handlers
	//------------------------------------------------------------
	/**
	 * GetLoginHandler - this is where we handle any GetLogin event
	 * requests. If they are already logged in, just redirect on.
	 */
	class GetLoginHandler extends DefaultBaseEventListener {
		public void handleControlEvent(ControlEventContext context) throws EventException, ServletException, IOException {
			//unpack the necessary entities from the context
			DispatchQueue queue = context.getQueue();

			//if it looks like we are already logged in, try and auto-login
			//otherwise, just render the login screen
			if (LoginServices.isLoggedIn(SessionServices.getSession(context.getRequest()))) {
				if (logger.isDebugEnabled()) logger.debug("Adding DoLogin to Queue (Auto-login)");
				queue.addEvent(new DoLogin(context.getEvent()));
			} else {
				if (logger.isDebugEnabled()) logger.debug("Adding RenderLogin to Queue");
				queue.addEvent(new RenderLogin(context.getEvent()));
			}
	    }
	}

	/**
	 * DoLoginHandler - this is where we handle any AttemptLogin 
	 * events. Validate and then redirect accordingly.
	 */
	class DoLoginHandler extends DefaultBaseEventListener {
		public void handleControlEvent(ControlEventContext context) throws EventException, ServletException, IOException {
			//unpack the necessary entities from the context
			HttpServletRequest req = context.getRequest();

			//create the login form
			ValidationException ve = null;
			if (logger.isDebugEnabled()) logger.debug("Creating login form");			
			LoginForm lf = new LoginForm(SessionServices.getSession(req));
			try {
				//map/validate the form
				if (logger.isDebugEnabled()) logger.debug("Mapping/Validating login form");			
				lf.map(req).validate(true);
			} catch (ValidationException e) {
				LoginServices.logOut(SessionServices.getSession(req));
				if (logger.isDebugEnabled()) logger.debug("Validation Exception: "+e);
				if (logger.isDebugEnabled()) Collections.printStackTrace(e.getExceptionList(), 0, logger, null);
				ve = e;
			}

			//store a copy of the form and any errors in the queue
			if (logger.isDebugEnabled()) logger.debug("Saving form, errors");			
			context.putState(LOGIN_FORM, lf);
			context.putState(LOGIN_ERR, ve);
			
			//redirect to the Main screen
			if (ve==null) {
				if (logger.isDebugEnabled()) logger.debug("Redirecting to GetCatalog");			
				throw new ClientSideRedirectException(new GetCatalog());
			} else {
				if (logger.isDebugEnabled()) logger.debug("InterruptDispatch to GetLogin");			
				throw new InterruptDispatchException(new GetLogin());
			}
	    }
	}

	/**
	 * ValidateLoginHandler - make sure the user is actually logged in
	 * if they try and fire any event that extends from ValidateEvent. If
	 * they're not logged in, redirect to the Login screen.
	 */
	class ValidateLoginHandler extends DefaultBaseEventListener {
		public void handleControlEvent(ControlEventContext context) throws EventException, ServletException, IOException {
			//if it looks like we are already logged in, try and auto-login
			//otherwise, just render the login screen
			if (!LoginServices.isLoggedIn(SessionServices.getSession(context.getRequest()))) {
				if (logger.isDebugEnabled()) logger.debug("InterruptDispatch to GetLogin");
				throw new ClientSideRedirectException(new GetLogin(context.getEvent()));
			}
	    }
	}

	/**
	 * DoLogoutHandler - this is where we handle any Logout requests
	 */
	class DoLogoutHandler extends DefaultBaseEventListener {
		public void handleControlEvent(ControlEventContext context) throws EventException, ServletException, IOException {
			//unpack the necessary entities from the context
			DispatchQueue queue = context.getQueue();

			//log the user out and forward control onto the logout screen
			if (logger.isDebugEnabled()) logger.debug("Logging out");
			LoginServices.logOut(SessionServices.getSession(context.getRequest()));
			queue.addEvent(new RenderLogout(context.getEvent()));
	    }
	}

	//------------------------------------------------------------
	//                Model 2 - View Event Handlers
	//------------------------------------------------------------
	/**
	 * RenderLoginHandler - 
	 */
	class RenderLoginHandler extends DefaultBaseEventListener {
		public void handleViewEvent(ViewEventContext context) throws EventException, ServletException, IOException {
			//unpack the necessary entities from the context
			HttpServletRequest req = context.getRequest();
org.enhydra.barracuda.core.util.http.ServletUtil.showParams(req, logger);

			//check to see if we need to generate a sample err (for demo purposes)
			if (logger.isDebugEnabled()) logger.debug("Handling RenderLogin Request");			
			if (new Boolean(req.getParameter("generateSampleError")).booleanValue()) {
				throw new InterruptDispatchException(new SampleError(new Exception("Look, here's an error!")));
			}

			//get the XMLC object
//			XMLCFactory factory = XMLCServices.getXMLCFactory(context.getConfig());
//			LoginHTML page = (LoginHTML) factory.create(LoginHTML.class);
            LoginHTML page = (LoginHTML) DefaultDOMLoader.getGlobalInstance().getDOM(LoginHTML.class);

//			// Swap in our real run-time JavaScript to replace the storyboard JavaScript
//			HTMLScriptElement script = new LoginScriptHTML().getElementRealScript();
//			XMLCUtil.replaceNode(script, page.getElementDummyScript());		

/*
			//repopulate the form (if possible)
			LoginForm lf = (LoginForm) context.get(LOGIN_FORM);
			if (lf!=null) {
				if (logger.isDebugEnabled()) logger.debug("Repopulating Form");			
	            page.getElementLogin().setValue(lf.getStringValDefault(LoginForm.USER, ""));
	            page.getElementPassword().setValue(lf.getStringValDefault(LoginForm.PASSWORD, ""));
			}

			//check for errors
			if (logger.isDebugEnabled()) logger.debug("Checking for errors");			
			ValidationException ve = (ValidationException) context.get(LOGIN_ERR);
			if (ve==null) {
				if (showDebug>2) Debug.println (this, 2, "No errors found");			
			    page.getElementErrorText().getParentNode().removeChild(page.getElementErrorText());
			} else {
				List errlist = ve.getExceptionList();
				if (showDebug>2) Debug.println (this, 2, errlist.size()+"error(s) found");			
				StringBuffer sb = new StringBuffer(errlist.size()>1 ? "There were several errors:" : "");
				Iterator it = errlist.iterator();
				while (it.hasNext()) {
					sb.append(" "+((ValidationException) it.next()).getMessage());
				}
			    page.setTextErrorText(sb.toString());
			}
*/

			//create a template component and render it
			if (logger.isDebugEnabled()) logger.debug("Creating template component");			
			Node node = page.getDocument().getElementById("LoginForm");
			BTemplate templateComp = new BTemplate(new LoginModel(context));
            templateComp.setView(new DefaultTemplateView(node));
			try {templateComp.render(new DefaultViewContext(context));} 
			catch (RenderException re) {logger.warn("Render err:"+re);}
			
			//now actually render the page (the DOMWriter will take care of actually
			//setting the content type)
			if (logger.isDebugEnabled()) logger.debug("Rendering document");			
			new DefaultDOMWriter().write(page, context.getResponse());
	    }
	}

	/**
	 * RenderLogoutHandler - 
	 */
/*     
	class RenderLogoutHandler extends DefaultBaseEventListener {
		public void handleViewEvent(ViewEventContext context) throws EventException, ServletException, IOException {
			//get the XMLC object
//			XMLCFactory factory = XMLCServices.getXMLCFactory(context.getConfig());
//			ExitHTML page = (ExitHTML) factory.create(ExitHTML.class);
            ExitHTML page = (ExitHTML) DefaultDOMLoader.getGlobalInstance().getDOM(ExitHTML.class);
			
			//now actually render it (the DOMWriter will take care of actually
			//setting the content type)
			if (logger.isDebugEnabled()) logger.debug("Rendering document");			
			new DefaultDOMWriter().write(page, context.getResponse());
	    }
	}
*/    
    class RenderLogoutHandler extends BTemplateViewHandler {
        //public TemplateModel getTemplateModel() {return new LogoutModel();}
        public Object getTemplateModels() {return new LogoutModel();}
        public Class getTemplateClass() {return ExitHTML.class;}
    }

	/**
	 * RenderSampleErrorHandler - render a SampleError screen
	 */
	class RenderSampleErrorHandler extends DefaultBaseEventListener {
		public void handleViewEvent(ViewEventContext context) throws EventException, ServletException, IOException {
			//unpack the necessary entities from the context
			BaseEvent event = context.getEvent();

			//get the XMLC object
//			XMLCFactory factory = XMLCServices.getXMLCFactory(context.getConfig());
//			ErrorHTML page = (ErrorHTML) factory.create(ErrorHTML.class);
            ErrorHTML page = (ErrorHTML) DefaultDOMLoader.getGlobalInstance().getDOM(ErrorHTML.class);
			Object source = event.getSource();
			if (source instanceof Throwable) {
				//get the error
				Throwable th = (Throwable) source;
	            StringWriter writer = new StringWriter();
	            th.printStackTrace(new PrintWriter(writer));
				String err = writer.toString();
				
				//log it
	            LogChannel logChannel = Enhydra.getLogChannel();
	            int level = logChannel.getLevel("DEBUG");
	            logChannel.write(level, this.getClass().getName()+" stack trace = ");
	            logChannel.write(level, err);
				
				//dump it to the error page as well
	            page.setTextStackTrace(err);
	            page.setTextErrorMessage(th.getMessage());
			}
        
			//now actually render it (the DOMWriter will take care of actually
			//setting the content type)
			if (logger.isDebugEnabled()) logger.debug("Rendering document");			
			new DefaultDOMWriter().write(page, context.getResponse());
	    }
	}

	//------------------------------------------------------------
	//               Template Models
	//------------------------------------------------------------
	/**
	 * LoginModel
	 */
	class LoginModel extends AbstractTemplateModel {
	
		LoginForm fm = null;
		ValidationException ve = null;

		//constructor (extract form, exception from context)		
		public LoginModel(EventContext ec) {
			fm = (LoginForm) ec.getState(LOGIN_FORM);
			ve = (ValidationException) ec.getState(LOGIN_ERR);
		}
	
		//register the model by name
		public String getName() {return "Login";}
		
		//provide items by key
        public Object getItem(String key) {
            ViewContext vc = getViewContext();
			if (key.equals("Username")) {
				return (fm!=null ? fm.getStringVal(LoginForm.USER, "") : "");
			} else if (key.equals("Password")) {
				return (fm!=null ? fm.getStringVal(LoginForm.PASSWORD, "") : "");
			} else if (key.equals("LoginButton")) {
                BAction baComp = new BAction(new DoLogin());
                baComp.setDisableBackButton(true);
				return baComp;
			} else if (key.equals("RegisterButton")) {
                BAction baComp = new BAction(new GetRegister());
                baComp.setDisableBackButton(true);
				return baComp;
			} else if (key.equals("ExceptionButton")) {
                BAction baComp = new BAction(new GetLogin());
                baComp.setParam("generateSampleError","true");
                baComp.setDisableBackButton(true);
				return baComp;
			} else if (key.equals("Errors")) {
//				if (ve==null) return "";
				if (ve==null) return null;
				List errlist = ve.getExceptionList();
				StringBuffer sb = new StringBuffer(errlist.size()>1 ? "There were several errors:" : "");
				Iterator it = errlist.iterator();
				while (it.hasNext()) {
					sb.append(" "+((ValidationException) it.next()).getMessage());
				}
			    return sb.toString();
			} else {
                return super.getItem(key);
			}
		}
	}

	/**
	 * LogoutModel
	 */
	class LogoutModel extends AbstractTemplateModel {
	
		//register the model by name
		public String getName() {return "Logout";}
		
		//provide items by key
        public Object getItem(String key) {
            ViewContext vc = getViewContext();
			if (key.equals("LoginButton")) {
                BAction baComp = new BAction(new GetLogin());
                baComp.setDisableBackButton(true);
				return baComp;
			} else {
                return super.getItem(key);
			}
		}
	}


	//------------------------------------------------------------
	//               HTML Form Mappings, Validators
	//------------------------------------------------------------
	/**
	 * Login form - define the login form
	 */
	class LoginForm extends DefaultFormMap {
		//login form constants (these values correspond to the HTML param names)
		static final String USER = "login";					
		static final String PASSWORD = "password";			

		HttpSession session = null;
		
		public LoginForm(HttpSession isession) {
			session = isession;

			//define the elements
			if (logger.isDebugEnabled()) logger.debug("Defining Login form elements");			
			this.defineElement(new DefaultFormElement(USER, 		FormType.STRING, null, new NotNullValidator("You must enter a Login.")));
			this.defineElement(new DefaultFormElement(PASSWORD, 	FormType.STRING, null, new NotNullValidator("You must enter a Password.")));
			
			//define a form validator
			if (logger.isDebugEnabled()) logger.debug("Defining Registration form validator");			
			this.defineValidator(new LoginValidator(session));
		}

		/**
		 * we override this method because instead of using the default validation
		 * process (validate elements, then form), we don't want to validate at all
		 * if they are already logged in...we just assume they're valid and return.
		 */
		public FormMap validate(boolean deferExceptions) throws ValidationException {
			//see if they are already logged in. If so, just consider them valid			
			if (logger.isDebugEnabled()) logger.debug("Checking to see if already logged in (prevalidated)");			
			if (LoginServices.isLoggedIn(session)) {
				try {
					Person p = LoginServices.getCurrentUser(session);
					String fuser = this.getStringVal(LoginForm.USER);
					String fpassword = this.getStringVal(LoginForm.PASSWORD);
					if (fuser==null && fpassword==null) return this;	//if user/pwd = null, just use the current settings
					if (p.getLogin().equals(fuser) &&
						p.getPassword().equals(fpassword)) return this;	//if they're != null, make sure they match, otherwise, we'll want to force a complete revalidate
				} catch (DiscRackBusinessException e) {}
			}
			
			//if not, go ahead and validate
			return super.validate(deferExceptions);
		}
	}

	/**
	 * Login validator - define a custom form validator
	 */
	class LoginValidator extends DefaultFormValidator {

		HttpSession session = null;
		
		public LoginValidator(HttpSession isession) {
			session = isession;
		}
		
		public void validateForm(FormMap imap, boolean deferExceptions) throws ValidationException {
			if (logger.isDebugEnabled()) logger.debug("Validating login form");			

			//validate the login information
			if (logger.isDebugEnabled()) logger.debug("Validate the user/pwd info");			
			LoginForm map = (LoginForm) imap;
			String user = map.getStringVal(LoginForm.USER);
			String password = map.getStringVal(LoginForm.PASSWORD);
			String errMsg = "Invalid username or password. Please check your information and try again.";
	        try {
				//make sure the login exists and the passwords match
	            Person person = PersonFactory.findPerson(user);
	            if (person==null) throw new ValidationException(map.getElement(LoginForm.USER), errMsg);
				if (!person.getPassword().equals(password)) throw new ValidationException(map.getElement(LoginForm.PASSWORD), errMsg);
				
				//log them in
				LoginServices.logIn(session, person);			
	        } catch (DiscRackBusinessException e) {
				LoginServices.logOut(session);	//just to be sure we don't get stuck in a loop
	            throw new ValidationException(map.getElement(LoginForm.USER), "System error finding user: " + e.getMessage()+". Please try again. If the problem persists, contact your system administrator.", e);
	        }
		}
	}
}

